"""
This module provides functions and classes to work with processes and sub-processes.

We combine two great packages, the `subprocess` and the `psutils` packages, to
provide a simplified, robust and user-friendly interface to work with sub-processes.
The classes and functions are optimized to work with processes within the framework
of the `cgse`, so we do not intend to be fully generic. If you need that, we recommend
to use the `subprocess` and `psutil` packages directly.

The main class provided is the `SubProcess` which by default, starts a sub-process
in the background and detached from the parent process. That means there is no
communication between the parent and the subprocess through pipes. Most (if not all)
processes in the `cgse` framework communicate with ZeroMQ messages in different
protocols like REQ-REP, PUSH-PULL and ROUTER-DEALER

Another useful class is the `ProcessStatus`. This class provides status information
like memory and CPU usage for the running process. Additionally, it will generate
and update metrics that can be queried by the Prometheus timeseries database.

"""
from __future__ import annotations

__all__ = [
    "get_process_info",
    "is_process_running",
    "list_processes",
    "list_zombies",
    "ps_egrep",
    "ProcessStatus",
    "SubProcess",
]

import contextlib
import datetime
import logging
import os
import subprocess
import sys
import threading
import time
import uuid
from typing import List
from typing import Optional

import psutil
from prometheus_client import Gauge

from egse.bits import humanize_bytes
from egse.system import humanize_seconds

_logger = logging.getLogger(__name__)


class ProcessStatus:
    """
    The ProcessStatus is basically a dataclass that contains the status information of a running
    process.

    The available information is the following:

    * pid: the process identifier
    * uptime: the process up-time as a floating point number expressed in seconds
    * uuid: the UUID1 for this process
    * memory info: memory information on the process
    * cpu usage, percentage and count (number of physical cores)

    The status will always be updated before returning or printing.

    Parameters:
        metrics_prefix: the prefix that identifies the process for which these metrics are gathered.
    """

    def __init__(self, *, metrics_prefix: Optional[str] = None):
        self._process = psutil.Process()
        self._cpu_count = psutil.cpu_count(logical=False)
        with self._process.oneshot():
            self._pid: int = self._process.pid
            self._create_time: float = self._process.create_time()
            # not sure if we need to use interval=0.1 as an argument in the next call
            self._cpu_percent: float = self._process.cpu_percent()
            self._cpu_times = self._process.cpu_times()
            self._uptime = (
                    datetime.datetime.now(tz=datetime.timezone.utc).timestamp() - self._create_time
            )
            self._memory_info = self._process.memory_full_info()
        self._uuid: uuid.UUID = uuid.uuid1()

        metrics_prefix = f"{metrics_prefix.lower()}_" if metrics_prefix else ""

        self.metrics = dict(
            PSUTIL_NUMBER_OF_CPU=Gauge(
                f"{metrics_prefix}psutil_number_of_cpu",
                "Number of physical cores, excluding hyper thread CPUs"
            ),
            PSUTIL_CPU_TIMES=Gauge(
                f"{metrics_prefix}psutil_cpu_times_seconds",
                "Accumulated process time in seconds", ["type"]
            ),
            PSUTIL_CPU_PERCENT=Gauge(
                f"{metrics_prefix}psutil_cpu_percent",
                "The current process CPU utilization as a percentage"
            ),
            PSUTIL_PID=Gauge(
                f"{metrics_prefix}psutil_pid", "Process ID"
            ),
            PSUTIL_MEMORY_INFO=Gauge(
                f"{metrics_prefix}psutil_memory_info_bytes",
                "Memory info for this instrumented process",
                ["type"]
            ),
            PSUTIL_NUMBER_OF_THREADS=Gauge(
                f"{metrics_prefix}psutil_number_of_threads",
                "Return the number of Thread objects currently alive"
            ),
            PSUTIL_PROC_UPTIME=Gauge(
                f"{metrics_prefix}psutil_process_uptime",
                "Return the time in seconds that the process is up and running"
            ),
        )

        self.metrics["PSUTIL_NUMBER_OF_CPU"].set(self._cpu_count)
        self.metrics["PSUTIL_PID"].set(self._process.pid)

        self.update()

    def update_metrics(self):
        """
        Updates the metrics that are taken from the psutils module.

        The following metrics are never updated since they are not changed during a
        process execution:

          * PSUTIL_NUMBER_OF_CPU
          * PSUTIL_PID
        """

        self.metrics["PSUTIL_MEMORY_INFO"].labels(type="rss").set(self._memory_info.rss)
        self.metrics["PSUTIL_MEMORY_INFO"].labels(type="uss").set(self._memory_info.uss)
        self.metrics["PSUTIL_CPU_TIMES"].labels(type="user").set(self._cpu_times.user)
        self.metrics["PSUTIL_CPU_TIMES"].labels(type="system").set(self._cpu_times.system)
        self.metrics["PSUTIL_CPU_PERCENT"].set(self._cpu_percent)
        self.metrics["PSUTIL_NUMBER_OF_THREADS"].set(threading.active_count())
        self.metrics["PSUTIL_PROC_UPTIME"].set(self._uptime)

    def update(self) -> ProcessStatus:
        """
        Updates those values that change during execution, like memory usage, number of
        connections, ...

        This call will also update the metrics!

        Returns:
            the ProcessStatus object, self.
        """
        self._cpu_percent = self._process.cpu_percent()
        self._cpu_times = self._process.cpu_times()
        self._uptime = time.time() - self._create_time
        self._memory_info = self._process.memory_full_info()

        self.update_metrics()

        return self

    def as_dict(self) -> dict:
        """Returns all process information as a dictionary.

        This runs the `update()` method first to bring the numbers up-to-date.
        """
        self.update()
        return {
            "PID": self._pid,
            "Up": self._uptime,
            "UUID": self._uuid,
            "RSS": self._memory_info.rss,
            "USS": self._memory_info.uss,
            "CPU User": self._cpu_times.user,
            "CPU System": self._cpu_times.system,
            "CPU count": self._cpu_count,
            "CPU%": self._cpu_percent,
        }

    def __str__(self):
        self.update()
        msg = (
            f"PID: {self._pid}, "
            f"Up: {humanize_seconds(self._uptime)}, "
            f"UUID: {self._uuid}, "
            f"RSS: {humanize_bytes(self._memory_info.rss)}, "
            f"USS: {humanize_bytes(self._memory_info.uss)}, "
            f"CPU User: {humanize_seconds(self._cpu_times.user)}, "
            f"CPU System: {humanize_seconds(self._cpu_times.system)}, "
            f"CPU Count: {self._cpu_count}, "
            f"CPU%: {self._cpu_percent}"
        )
        return msg


#  * can we restart the same sub process?
#  * do we need to pass the additional arguments to the constructor or to the execute method?
#    When we can restart/re-execute a subprocess, we might want to do that with additional
#    arguments, e.g. to set a debugging flag or to start in simulator mode. Then we will need to
#    do that in the execute method.
#  * Process should have a notion of UUID, which it can request at start-up to communicate to the
#    process manager which can then check if it's known already or a new process that was started
#    (possible on another computer)


class SubProcess:
    """
    A SubProcess that is usually started by the ProcessManager.

    Example:

        proc = SubProcess("MyApp", [sys.executable, "-m", "egse.<module>"])
        proc.execute()



    """

    def __init__(
            self, name: str, cmd: List, args: List = None, shell: bool = False,
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    ):
        self._popen = None
        self._sub_process: psutil.Process | None = None
        self._name = name
        self._cmd = [str(x) for x in cmd]
        self._args = [str(x) for x in args] if args else []
        self._shell = shell
        self._stdout = stdout
        self._stderr = stderr

        self._exc_info = {}

    def execute(self) -> bool:
        """
        Execute the sub-process.

        Returns:
            True if the process could be started, False on error.
        """
        self._exc_info = {}

        try:
            command: List | str = [*self._cmd, *self._args]
            if self._shell:
                command = " ".join(command)
            _logger.debug(f"SubProcess command: {command}")
            self._popen = subprocess.Popen(
                command,
                env=os.environ,
                shell=self._shell,  # executable='/bin/bash',
                stdout=self._stdout,
                stderr=self._stderr,
                stdin=subprocess.DEVNULL,
            )
            self._sub_process = psutil.Process(self._popen.pid)

            _logger.debug(
                f"SubProcess started: {command}, pid={self._popen.pid}, sub_process="
                f"{self._sub_process} [pid={self._sub_process.pid}]"
            )
        except Exception as exc:
            # This error is raised when the command is not an executable or is not found
            _logger.error(f"Could not execute sub-process: {exc}", exc_info=True)
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self._exc_info = {
                'exc_type': exc_type,
                'exc_value': exc_value,
                'exc_traceback': exc_traceback,
                'command': " ".join([*self._cmd, *self._args]),
            }
            return False
        return True

    @property
    def name(self):
        return self._name

    @property
    def pid(self) -> int:
        return self._sub_process.pid if self._sub_process else None

    @property
    def exc_info(self) -> dict:
        return self._exc_info

    def cmdline(self) -> str:
        return " ".join(self._sub_process.cmdline())

    def children(self, recursive: bool = True) -> List:
        return self._sub_process.children(recursive=recursive)

    def is_child(self, pid: int):
        return any(pid == p.pid for p in self._sub_process.children(recursive=True))

    def is_running(self) -> bool:
        """
        Check if this process is still running.

        * checks if process exists
        * checks if process is not a zombie and is not dead

        Returns:
            True if the process is running.
        """
        if self._sub_process is None:
            return False
        if self._sub_process.is_running():
            # it still might be a zombie process
            if self._sub_process.status() in [psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD]:
                _logger.warning("The sub-process is dead or a zombie.")
                return False
            return True
        # _logger.debug(f"Return value of the sub-process: {self._popen.returncode}")
        return False

    def is_dead_or_zombie(self):
        if self._sub_process is None:
            return False
        if self._sub_process.is_running():
            # it might be a zombie process
            if self._sub_process.status() in [psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD]:
                return True
        return False

    def exists(self) -> bool:
        """
        Checks if the sub-process exists by checking if its process ID exists.

        Returns:
            True if the sub-process exists.
        """
        return psutil.pid_exists(self.pid)

    def quit(self) -> int:
        """
        Send a request to quit to the process.

        This will first send a SIGTERM signal to the process, if that fails,
        a SIGKILL will be sent.

        Returns:
            0 if the process and its sub-processes are all terminated. Will
                return > 0 to indicate the number of processes that survived the
                SIGKILL.
        """
        return self.reap_children()

    def reap_children(self, timeout=3) -> int:
        """
        Tries hard to terminate and ultimately kill all the children of this process.

        This will first send a SIGTERM signal to the process, if that fails,
        a SIGKILL will be sent.

        Returns:
            0 if the process and its sub-processes are all terminated. Will
                return > 0 to indicate the number of processes that survived the
                SIGKILL.
        """

        def on_terminate(proc):
            _logger.info(f"process {proc} terminated with exit code {proc.returncode}")

        return_code = 0
        self._exc_info = {}

        children = self._sub_process.children()

        _logger.info(f"Children: {children}")

        # send SIGTERM to subprocess

        try:
            _logger.info(f"Send a SIGTERM to process with PID={self.pid}")
            self._sub_process.terminate()
            try:
                return_code = self._sub_process.wait(timeout=5.0)  # make this timeout an instance parameter
                _logger.info(f"{return_code = }")
            except psutil.TimeoutExpired:
                _logger.info(f"TimeoutExpired after 5s for PID={self.pid}")
                _logger.info(f"Send a SIGKILL to process with PID={self.pid}")

                exc_type, exc_value, exc_traceback = sys.exc_info()
                self._exc_info = {
                    'exc_type': exc_type,
                    'exc_value': exc_value,
                    'exc_traceback': exc_traceback,
                    'command': " ".join([*self._cmd, *self._args]),
                }

                self._sub_process.kill()
                return -9  # meaning the process was terminated by a SIGKILL
        except psutil.NoSuchProcess:
            # If we get here, the process died already and there are also no children to terminate
            _logger.info(f"NoSuchProcess with PID={self._sub_process.pid}")

            exc_type, exc_value, exc_traceback = sys.exc_info()
            self._exc_info = {
                'exc_type': exc_type,
                'exc_value': exc_value,
                'exc_traceback': exc_traceback,
                'command': " ".join([*self._cmd, *self._args]),
            }

            return 0

        # now terminate the children

        if children:

            for p in children:
                try:
                    _logger.info(f"Send a SIGTERM to child process with PID={p.pid}")
                    p.terminate()
                except psutil.NoSuchProcess:
                    pass
            gone, alive = psutil.wait_procs(children, timeout=timeout, callback=on_terminate)
            if alive:
                # send SIGKILL
                for p in alive:
                    _logger.info(f"Child process {p} survived SIGTERM; trying SIGKILL")
                    try:
                        _logger.info(f"Send a SIGKILL to child process with PID={p.pid}")
                        p.kill()
                    except psutil.NoSuchProcess:
                        pass
                gone, alive = psutil.wait_procs(alive, timeout=timeout, callback=on_terminate)
                if alive:
                    # give up
                    for p in alive:
                        _logger.info(f"Child process {p} survived SIGKILL; giving up")

        return return_code

    def returncode(self):
        """
        Check if the sub-process is terminated and return its return code or None when the process
        is still running.
        """
        return self._popen.poll() if self._popen else None

    def communicate(self) -> tuple[str, str]:
        output, error = self._popen.communicate()
        return output.decode() if output else None, error.decode() if error else None


def list_processes(
        items: List[str] | str, contains: bool = True, case_sensitive: bool = False, verbose: bool = False
) -> list[dict]:
    """
    Returns and optionally prints the processes that match the given criteria in items.

    Args:
        items: a string or a list of strings that should match command line parts
        contains: if True, the match is done with 'in' otherwise '==' [default: True]
        case_sensitive: if True, the match shall be case-sensitive [default: False]
        verbose: if True, the processes will be printed to the console

    Returns:
        A list of dictionaries for the matching processes. The dict contains the
            'pid', 'status' and 'cmdline' of a process.
    """
    procs = is_process_running(items, contains=contains, case_sensitive=case_sensitive, as_list=True)

    result = []

    for pid in procs:
        proc = psutil.Process(pid)
        status = proc.status()
        cmdline = ' '.join(proc.cmdline())
        result.append({'pid': pid, 'status': status, 'cmdline': cmdline})

    if verbose:
        if result:
            print(f"{'PID':5s} {'Status':>20s} {'Commandline'}")
            print(
                "\n".join(
                    [
                        f"{entry['pid']:5d} {entry['status']:>20s} {entry['cmdline']}"
                        for entry in result
                    ]
                ))
        else:
            print(f"No processes found for {items}.")

    return result


def list_zombies():
    """
    Returns a list of zombie processes.

    A zombie process, also known as a defunct process, is a process that has
    completed its execution but still has an entry in the process table.
    This happens when a child process terminates, but the parent process hasn't
    yet read its exit status by using a system call like wait(). As a result,
    the process is "dead" (it has completed execution) but hasn't been "reaped"
    or removed from the system's process table.

    A zombie process can not be killed with SIGKILL because it's already dead, and
    it's only removed when the parent process reads their exit status or when the
    parent process itself terminates.

    A zombie process does not block ports, so it's mostly harmless and will disappear
    when the parent process terminates.

    Returns:
        A list of dictionaries with information on the zombie processes. The dict
            contains the 'pid', 'name', and 'cmdline' of the zombie process.
    """
    zombies = []

    for proc in psutil.process_iter(['pid', 'name', 'status', 'cmdline']):
        try:
            if proc.info['status'] == psutil.STATUS_ZOMBIE:
                zombies.append(
                    {
                        'pid': proc.info['pid'],
                        'name': proc.info['name'],
                        'cmdline': proc.info['cmdline'],
                    }
                )
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass

    return zombies


def is_process_running(
        items: List[str] | str,
        contains: bool = True, case_sensitive: bool = False, as_list: bool = False) -> (int | List[int]):
    """
    Check if there is any running process that contains the given items in its
    commandline.

    Loops over all running processes and tries to match all items in the 'items'
    argument to the command line of the process. If all 'items' can be matched
    to a process, the function returns the PID of that process.

    By default, only the first matching process PID is returned. If `as_list=True`
    then all mathing process PIDs are returned as a list.

    Args:
        items: a string or a list of strings that should match command line parts
        contains: if True, the match is done with 'in' otherwise '==' [default: True]
        case_sensitive: if True, the match shall be case-sensitive [default: False]
        as_list: return the PID of all matching processes as a list [default: False]

    Returns:
        The PID(s) if there exists a running process with the given items, 0 or [] otherwise.
    """

    def lower(x: str) -> str:
        return x.lower()

    def pass_through(x: str) -> str:
        return x

    case = pass_through if case_sensitive else lower

    if not items:
        _logger.warning("Expected at least one item in 'items', none were given. False returned.")
        return [] if as_list else 0

    items = [items] if isinstance(items, str) else items

    found = []

    for proc in psutil.process_iter(attrs=['pid', 'cmdline', 'name'], ad_value='n/a'):
        with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            # _logger.info(f"{proc.name().lower() = }, {proc.cmdline() = }")
            if contains:
                if all(any(case(y) in case(x) for x in proc.cmdline()) for y in items):
                    found.append(proc.pid)
            elif all(any(case(y) == case(x) for x in proc.cmdline()) for y in items):
                found.append(proc.pid)
    if found:
        return found if as_list else found[0]
    else:
        return [] if as_list else 0


def get_process_info(items: List[str] | str, contains: bool = True, case_sensitive: bool = False) -> List:
    """
    Loops over all running processes and tries to match each item in 'items' to the command line
    of the process. Any process where all 'items' can be matched will end up in the response.

    Returns a list with the process info (PID, cmdline, create_time) for any processes where all 'items' match
    the process command line. An empty list is returned when not 'all the items' match for any of the
    processes.

    Examples:
        >>> get_process_info(items=["feesim"])
        [
            {
                'pid': 10166,
                'cmdline': [
                    '/Library/Frameworks/Python.framework/Versions/3.8/Resources/Python.app/Contents/MacOS/Python',
                    '/Users/rik/git/plato-common-egse/venv38/bin/feesim',
                    'start',
                    '--zeromq'
                ],
                'create_time': 1664898231.915995
            }
        ]

        >>> get_process_info(items=["dpu_cs", "--zeromq"])
        [
            {
                'pid': 11595,
                'cmdline': [
                    '/Library/Frameworks/Python.framework/Versions/3.8/Resources/Python.app/Contents/MacOS/Python',
                    '/Users/rik/git/plato-common-egse/venv38/bin/dpu_cs',
                    'start',
                    '--zeromq'
                ],
                'create_time': 1664898973.542281
            }
        ]

    Args:
        items: a string or a list of strings that should match command line items
        contains: if True, the match is done with 'in' otherwise '=='
        case_sensitive: if True, the match shall be case-sensitive

    Returns:
        A list of process info entries.

    """
    response = []

    def lower(x: str) -> str:
        return x.lower()

    def pass_through(x: str) -> str:
        return x

    case = pass_through if case_sensitive else lower

    if not items:
        _logger.warning("Expected at least one item in 'items', none were given. Empty list returned.")
        return response

    items = [items] if isinstance(items, str) else items

    for proc in psutil.process_iter():
        with contextlib.suppress(psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            # _logger.info(f"{proc.name().lower() = }, {proc.cmdline() = }")
            if contains:
                if all(any(case(y) in case(x) for x in proc.cmdline()) for y in items):
                    response.append(proc.as_dict(attrs=['pid', 'cmdline', 'create_time']))
            elif all(any(case(y) == case(x) for x in proc.cmdline()) for y in items):
                response.append(proc.as_dict(attrs=['pid', 'cmdline', 'create_time']))

    return response


def ps_egrep(pattern):
    # First command-line
    ps_command = ["ps", "-ef"]

    # Second command-line
    grep_command = ["egrep", pattern]

    # Launch first process
    ps_process = subprocess.Popen(ps_command, stdout=subprocess.PIPE)

    # Launch second process and connect it to the first one
    grep_process = subprocess.Popen(
        grep_command, stdin=ps_process.stdout, stdout=subprocess.PIPE
    )

    # Let stream flow between them
    output, _ = grep_process.communicate()

    response = [
        line
        for line in output.decode().rstrip().split('\n')
        if line and "egrep " not in line
    ]

    return response



from egse.hexapod.symetrie import ControllerFactory
from egse.hexapod.symetrie import ProxyFactory
from egse.hexapod.symetrie import get_hexapod_controller_pars
from egse.hexapod.symetrie.dynalpha import AlphaPlusControllerInterface
from egse.hexapod.symetrie.dynalpha import AlphaPlusTelnetInterface
from egse.mixin import DynamicCommandMixin
from egse.proxy import DynamicProxy
from egse.settings import Settings
from egse.zmq_ser import connect_address

CTRL_SETTINGS = Settings.load("Hexapod PUNA Control Server")


class PunaPlusInterface(AlphaPlusControllerInterface):
    """
    Interface definition for the PunaPlusController and the PunaPlusProxy.
    """


class PunaPlusController(PunaPlusInterface, DynamicCommandMixin):
    def __init__(self, hostname: str = "127.0.0.1", port: int = 23):
        self.transport = self.device = AlphaPlusTelnetInterface(hostname, port)
        self.hostname = hostname
        self.port = port

        super().__init__()

    def get_controller_type(self):
        return "ALPHA+"

    def is_simulator(self):
        return False

    def is_connected(self):
        return self.device.is_connected()

    def connect(self):
        self.device.connect()

    def disconnect(self):
        self.device.disconnect()

    def reconnect(self):
        if self.is_connected():
            self.disconnect()
        self.connect()


class PunaPlusProxy(DynamicProxy, PunaPlusInterface):
    """
    The PunaPlusProxy class is used to connect to the control server and send commands to the
    Hexapod PUNA remotely. The devce controller for that PUNA hexapod is an Alpha+ controller.
    """

    def __init__(
        self,
        protocol=CTRL_SETTINGS.PROTOCOL,
        hostname=CTRL_SETTINGS.HOSTNAME,
        port=CTRL_SETTINGS.COMMANDING_PORT,
    ):
        """
        Args:
            protocol: the transport protocol [default is taken from settings file]
            hostname: location of the control server (IP address) [default is taken from settings
            file]
            port: TCP port on which the control server is listening for commands [default is
            taken from settings file]
        """
        super().__init__(connect_address(protocol, hostname, port))


if __name__ == "__main__":

    from egse.hexapod.symetrie.puna import PunaProxy

    # The following imports are needed for the isinstance() to work
    from egse.hexapod.symetrie.punaplus import PunaPlusProxy
    from egse.hexapod.symetrie.punaplus import PunaPlusController

    print()

    *_, device_id, device_name, _ = get_hexapod_controller_pars()
    print(f"{device_name = }, {device_id = }")

    factory = ProxyFactory()
    proxy = factory.create(device_name, device_id="1A")
    assert isinstance(proxy, PunaProxy)

    proxy = factory.create(device_name, device_id="2B")
    assert isinstance(proxy, PunaPlusProxy)

    print(proxy.info())

    factory = ControllerFactory()

    device = factory.create("PUNA", device_id="H_2B")
    device.connect()
    assert isinstance(device, PunaPlusController)

    print(device.info())

    device = factory.create("ZONDA")
    device.connect()

    print(device.info())

    device.disconnect()



"""
This module defines Listeners between control servers and is part of a notification system where changes in
one control-server are notified to all listeners of that control-server.

Since we have communication between control servers, a listener registers to a control server with its name/id
and a Proxy class that will be used to send the notification. A listener can also contain an `event_id` to restrict
events to handle. (Shall events be filtered on the server (only notify events that match the `event_id` or shall they
be filtered on the client (only handle those events that match the `event_id`) ?)

Any control server will support notification and listeners are added through the service proxy of the control server.

A control server that wants to be notified by an event needs to implement the
[EventInterface](listener.md#egse.listener.EventInterface) in the Proxy and in the Controller.

"""
import logging
from enum import IntEnum
from typing import Any
from typing import Dict
from typing import List
from typing import Type

from egse.decorators import dynamic_interface
from egse.system import Timer

LOGGER = logging.getLogger(__name__)


class EVENT_ID(IntEnum):  # noqa
    """An identifier for the type of event."""
    ALL = 0
    """Match all events."""
    SETUP = 1
    """An event for a new or updated Setup."""


class Event:
    """An event that is generated by a control server."""
    def __init__(self, event_id: int, context: Any):
        self.id = event_id
        self.context = context

    def __repr__(self):
        return f"Event({EVENT_ID(self.id).name}, {self.context})"

    @property
    def type(self) -> str:
        try:
            return self.context['event_type']
        except KeyError:
            return "unknown: event_type not provided"


class EventInterface:
    """
    A dynamic interface for handling events.

    This interface defines a single method, 'handle_event', which is intended
    to be implemented by classes that want to handle specific types of events.

    Use this interface as a mixin for classes (Proxy/Controller) that implement this `handle_event` method.

    """
    @dynamic_interface
    def handle_event(self, event: Event):
        """Handles the specified event.

        Args:
            event (Event): An instance of the Event class representing the event to be handled.
        """
        ...


class Listeners:
    """
    A class for managing and notifying registered listeners.

    This class provides methods to add, remove, and notify listeners of events.
    """

    def __init__(self):
        """Initializes an instance of the Listeners class."""
        self._listeners: Dict[str, dict] = {}

    def __len__(self):
        """Returns the number of registered listeners."""
        return len(self._listeners)

    def add_listener(self, listener: dict):
        """
        Adds a new listener to the registered listeners.

        The listener argument dictionary is expected to have at least the following key:values pairs:

        * 'name': the name or identifier of the listener
        * 'proxy': a Proxy object that will be used for notifying the service

        Args:
            listener (dict): A dictionary with properties of the listener,
                including 'name' and 'proxy'.

        Raises:
            ValueError: If the listener already exists.
        """
        try:
            listener_name = listener["name"]
        except KeyError as exc:
            raise ValueError(f"Expected 'name' key in listener argument {listener}.") from exc

        if listener_name in self._listeners:
            raise ValueError(f"Process {listener_name} is already registered as a listener.")

        # For now we make these listeners have a Proxy class that will be used for notification. Later on
        # we might have other mechanisms for notification.

        from egse.proxy import Proxy

        proxy = listener.get('proxy')
        if proxy is not None:
            if not isinstance(proxy, type) or not issubclass(proxy, Proxy):
                raise ValueError(f"Expected 'proxy' in listener argument {proxy=} to be a Proxy sub-class.")

        self._listeners[listener_name] = listener

    def remove_listener(self, listener: dict):
        """
        Removes a listener from the registered listeners.

        Args:
           listener (dict): A dictionary representing the listener to be
               removed. It should contain a 'name' key.

        Raises:
           ValueError: If the 'name' key is not present in the listener
               argument or if the specified listener is not registered.
        """

        try:
            listener_name = listener["name"]
        except KeyError as exc:
            raise ValueError(f"Expected 'name' key in listener argument {listener}.") from exc

        try:
            del self._listeners[listener_name]
        except KeyError as exc:
            raise ValueError(f"Process {listener_name} cannot be removed, not registered.") from exc

    def notify_listeners(self, event: Event):
        """
        Notifies all registered listeners of a specific event.

        Args:
            event (Event): An instance of the Event class representing the
                event to be broadcasted.

        Note:
            The 'handle_event' method is called on each listener's proxy
            object to process the event.
        """

        LOGGER.debug(f"Notifying listeners {self.get_listener_names()}")

        for name, listener in self._listeners.items():
            self.notify_listener(name, listener, event)

    def notify_listener(self, name: str, listener: dict, event: Event):
        """
        Notifies a registered listener fro the given event.
        """
        proxy: Type[EventInterface] = listener['proxy']

        # The Proxy uses a REQ-REP protocol and waits for an answer. That's why we have a Timer
        # in the context so we are able to monitor how long it takes to notify a listener.
        # The Timer can be removed when we are confident all `handle_event` functions are properly
        # implemented.
        with Timer(f"Notify listener {name}", log_level=logging.DEBUG), proxy() as pobj:
            rc = pobj.handle_event(event)

        LOGGER.debug(f"Listener {name} returned {rc=}")

    def get_listener_names(self) -> List[str]:
        """Returns a list with the names of the registered listeners."""
        return list(self._listeners.keys())



import abc
import logging
import textwrap
from pathlib import Path
from typing import Optional

from egse.decorators import borg
from egse.decorators import deprecate
from egse.setup import Setup

LOGGER = logging.getLogger(__name__)


class StateError(Exception):
    pass


class UnknownStateError(StateError):
    pass


class IllegalStateTransition(StateError):
    pass


class NotImplementedTransition(StateError):
    pass


class ConnectionStateInterface(abc.ABC):
    """
    A class used to enforce the implementation of the _connection_ interface
    to model the state of a (network) connection.

    Subclasses only need to implement those methods that are applicable to
    their state.

    """

    # This class is to enforce the implementation of the interface on both the
    # model, i.e. the Proxy, and the State class. At the same time, it will allow
    # the State subclasses to implement only those methods that are applicable
    # in their state.

    @abc.abstractmethod
    def connect(self, proxy):
        pass

    @abc.abstractmethod
    def disconnect(self, proxy):
        pass

    @abc.abstractmethod
    def reconnect(self, proxy):
        pass


@borg
class _GlobalState:
    """
    This class implements global state that is shared between instances of this class.
    """

    # TODO (rik): turn command sequence into a class and move add_, clear_ and get_ to that class

    def __init__(self):
        self._dry_run = False
        self._command_sequence = []
        self._setup: Optional[Setup] = None

    def __call__(self, *args, **kwargs):
        return self

    @property
    def dry_run(self):
        return self._dry_run

    @dry_run.setter
    def dry_run(self, flag: bool):
        self._dry_run = flag

    def add_command(self, cmd):
        self._command_sequence.append(cmd)

    def get_command_sequence(self):
        return self._command_sequence

    def clear_command_sequence(self):
        self._command_sequence.clear()

    @property
    def setup(self) -> Optional[Setup]:
        """
        Returns the currently active Setup from the configuration manager. Please note that each call
        to this property sends a request to the configuration manager to return its current Setup. If
        you are accessing information from the Setup in a loop or function that is called often, save
        the Setup into a local variable before proceeding.

        Returns:
            The currently active Setup or None (when the configuration manager is not reachable).
        """
        return self.load_setup()

    # This function should be the standard function to reload a setup from the configuration manager
    # Since we have no proper CM yet, the function loads from the default setup.yaml file.
    # But what happens then in other parts of the system, where e.g. the PM has also the 'rights'
    # to call load_setup() on the CM_CS?

    def load_setup(self) -> Optional[Setup]:
        """
        Loads the currently active  Setup from the Configuration manager. The current Setup is the Setup
        that is defined and loaded in the Configuration manager. When the configuration manager is not
        reachable, None will be returned and a warning will be logged.

        Since the GlobalState should reflect the configuration of the test, it can only load the current
        Setup from the configuration manager. If you need to work with different Setups, work with the `Setup`
        class and the Configuration Manager directly.

        Returns:
            The currently active Setup or None.
        """
        from egse.confman import ConfigurationManagerProxy
        from egse.confman import is_configuration_manager_active

        if is_configuration_manager_active():
            with ConfigurationManagerProxy() as cm_proxy:
                self._setup = cm_proxy.get_setup()
        else:
            LOGGER.warning(
                textwrap.dedent(
                    """\
                    Could not reach the Configuration Manager to request the Setup, returning the current local Setup.

                    Check if the Configuration Manager is running and why it can not be consulted. When it's
                    back on-line, do a 'load_setup()'.
                    """
                )
            )

        return self._setup

    # FIXME:
    #  These two 'private' methods are still called in plato-test-scripts, so until that is fixed,
    #  leave the methods here

    @deprecate(reason="this is a private function", alternative="reload_setup()")
    def _reload_setup(self):
        self._setup = Setup.from_yaml_file()
        return self._setup

    @deprecate(reason="this is a private function", alternative="reload_setup_from()")
    def _reload_setup_from(self, filename: Path):
        """Used by the unit tests to load a predefined setup."""
        self._setup = Setup.from_yaml_file(filename=filename)
        return self.setup


GlobalState = _GlobalState()

__all__ = [
    "GlobalState",
]

if __name__ == "__main__":

    from rich import print

    print(textwrap.dedent(
        f"""\
            GlobalState info:
              Setup loaded: {GlobalState.setup.get_id()}
              Dry run: {'ON' if GlobalState.dry_run else 'OFF'}
              Command Sequence: {GlobalState.get_command_sequence()} \
        """)
    )



import datetime
import textwrap
from pathlib import Path

import rich

from egse.env import get_site_id
from egse.hk import get_housekeeping
from egse.setup import Setup
from egse.system import EPOCH_1958_1970
from egse.system import format_datetime

from fixtures.helpers import create_text_file

HERE = Path(__file__).parent


def test_get_housekeeping(default_env):

    data_dir = Path(default_env.data_root)

    # The get_housekeeping() function will by default search for HK telemetry in the daily CSV file that contains the
    # HK parameter that is passed into the function. Since we have no operational system running during the tess,
    # we will need to create sample data for that parameter in the correct file. The file will be located in the
    # data storage location folder (defined by the default_env fixture) at
    # `HERE/data/LAB23/daily/YYYYMMDD/YYYYMMDD_LAB23_DAQ-TM.csv`, where 'LAB23' is the SITE_ID and 'DAQ-TM' is the
    # storage mnemonic of the device (this is read from the TM dictionary file).

    today = format_datetime("today")
    today_with_dash = format_datetime("today", fmt="%Y-%m-%d")

    tm_dictionary_path = data_dir / "../common/telemetry/tm-dictionary.csv"

    create_text_file(tm_dictionary_path, textwrap.dedent(
        """\
        TM source;Storage mnemonic;CAM EGSE mnemonic;Original name in EGSE;Name of corresponding timestamp;\
        Origin of synoptics at CSL;Origin of synoptics at CSL1;Origin of synoptics at CSL2;\
        Origin of synoptics at SRON;Origin of synoptics at IAS;Origin of synoptics at INTA;\
        Description;MON screen;unit cal1;offset b cal1;slope a cal1;calibration function;\
        MAX nonops;MIN nonops;MAX ops;MIN ops;Comment

        Unit Test Manager;DAQ-TM;TEMP_ABC_001;ABC_001;timestamp;;;;;;;Temperature of ABC;;;;;;;;;;

        """
    ), create_folder=True)

    hk_path = data_dir / f"daily/{today}/{today}_{get_site_id()}_DAQ-TM.csv"

    create_text_file(hk_path, textwrap.dedent(
            f"""\
            timestamp,TEMPT_ABC_000,TEMP_ABC_001,TEMP_ABC_002
            {today_with_dash}T00:00:23.324+0000,21.333,23.3421,26.234
            {today_with_dash}T00:00:42.123+0000,22.145,23.4567,27.333
            """
        ), create_folder=True
    )

    # The TM dictionary will be loaded relative from the configuration data location as defined by
    # `get_conf_data_location()` (the latter being defined from the data storage location which is defined
    # in `default_env`).

    setup = Setup.from_yaml_string(
        textwrap.dedent(
            """
            telemetry:
                dictionary: pandas//../../common/telemetry/tm-dictionary.csv
                separator: ;
            """
        )
    )

    rich.print(f"{hk_path = }")
    rich.print(f"{tm_dictionary_path = }")

    try:
        timestamp, data = get_housekeeping("TEMP_ABC_001", setup=setup)
        timestamp -= EPOCH_1958_1970
        dt = datetime.datetime.utcfromtimestamp(timestamp)

        rich.print(f"{timestamp}, {dt}, {data}")

        assert data.strip() == '23.4567'
        assert format_datetime(dt, fmt="%Y-%m-%d").startswith(today_with_dash)

    finally:
        # Get rid of the CSV file
        hk_path.unlink()

        # Get rid of the tm dictionary file
        tm_dictionary_path.unlink()


def test_convert_hk_names():

    a = {
        'aaa': 1,
        'bbb': 2,
        'ccc': 3,
        'eee': 4,
    }

    c = {
        'aaa': 'AAA',
        'bbb': 'BBB',
        'ccc': 'CCC',
        'ddd': 'DDD',
    }

    from egse.hk import convert_hk_names

    b = convert_hk_names(a, c)

    # Result:
    #  * all keys in 'a' that have a conversion in 'c' shall be in 'b' with the converted key
    #  * all keys in 'a' that do not have a conversion in 'c', shall be in 'b' with their original key
    #  * all conversion keys that are in 'c' but not in 'a' shall just be ignored

    assert 'AAA' in b
    assert 'BBB' in b
    assert 'CCC' in b
    assert 'eee' in b

    assert 'aaa' not in b
    assert 'bbb' not in b
    assert 'ccc' not in b
    assert 'ddd' not in b
    assert 'DDD' not in b

    for k, v in a.items():
        if k == 'eee':
            assert b[k] == v
        else:
            assert b[k.upper()] == v



import logging
import re
from pathlib import Path
from typing import Dict, Union
from typing import List
from typing import Tuple

from egse.decorators import dynamic_interface
from egse.device import DeviceConnectionState
from egse.device import DeviceInterface
from egse.mixin import DynamicCommandMixin, CommandType
from egse.mixin import add_lf
from egse.mixin import dynamic_command
from egse.proxy import Proxy
from egse.settings import Settings
from egse.tempcontrol.keithley.daq6510_devif import DAQ6510EthernetInterface
from egse.zmq_ser import connect_address

logger = logging.getLogger(__name__)

HERE = Path(__file__).parent

CTRL_SETTINGS = Settings.load("Keithley Control Server")
FW_SETTINGS = Settings.load("Keithley DAQ6510")
DEVICE_SETTINGS = Settings.load(location=HERE, filename="daq6510.yaml")


DEFAULT_BUFFER_1 = "defbuffer1"
DEFAULT_BUFFER_2 = "defbuffer2"


class DAQ6510Interface(DeviceInterface):
    """
    Interface definition for the Keithley DAQ6510 Controller, Proxy, and Simulator.
    """

    @dynamic_interface
    def send_command(self, command: str, response: bool) -> Union[None, str]:
        """ Sends the given SCPI command to the device.

        The valid commands are described in the DAQ6510 Reference Manual [DAQ6510-901-01 Rev. B / September 2019].

        Args:
            command (str): SCPI command as specified in the DAQ6510 Reference Manual
            response (bool): Indicates whether you expect a reply from the device

        Returns: Response from the DAQ6510 is returned when a response was expected. When `response` is False, None
                 will be returned.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="*IDN?",
        process_cmd_string=add_lf,
    )
    def info(self) -> str:
        """ Returns basic information about the device, its name, firmware version, etc.

        The string returned is subject to change without notice and can not be used for parsing information.

        Returns: Identification string of the instrument.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="*RST",
        process_cmd_string=add_lf,
    )
    def reset(self) -> None:
        """ Resets the DAQ6510.

        This returns the instrument to default settings, and cancels all pending commands.

        Note:
            The `reset()` method also deletes all the user-defined buffers. The two default buffers are cleared.
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.WRITE,
                     cmd_string=":SYST:TIME ${year}, ${month}, ${day}, ${hour}, ${minute}, ${second}")
    def set_time(self, year: int, month: int, day: int, hour: int, minute: int, second: int) -> None:
        """ Sets the absolute date and time for the device.

        Args:
            year (int): Year
            month (int): Month
            day (int): Day
            hour (int): Hour
            minute (int): Minute
            second (int): Second
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string=":SYST:TIME? 1")
    def get_time(self) -> str:
        """ Gets the date and time from the device in UTC.

        The returned string is of the format:

            <weekday> <month> <day> <hour>:<minute>:<second> <year>

        Returns: Data and time from the device in UTC.
        """

        raise NotImplementedError

    @dynamic_interface
    def read_buffer(self, start: int, end: int, buffer_name: str, elements: List[str]):
        """ Reads specific data elements (measurements) from the given buffer.

        Args:
            start: (int) First index of the buffer that should be returned (>= 1)
            end (int): Last index of the buffer that should be returned
            buffer_name (str): Name of the buffer to read out
            elements (List[str]): List of elements from the buffer to include in the response

        Returns: List of all the readings.
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="TRAC:ACTUAL? ${buffer_name}")
    def get_buffer_count(self, buffer_name: str = DEFAULT_BUFFER_1):
        """ Returns the number of data points in the specified buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="TRACE:POINTS? ${buffer_name}")
    def get_buffer_capacity(self, buffer_name: str):
        """ Returns the capacity of the specified buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.WRITE, cmd_string="TRACE:DELETE ${buffer_name}")
    def delete_buffer(self, buffer_name: str) -> None:
        """ Deletes the specified buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        raise NotImplementedError

    @dynamic_interface
    def clear_buffer(self, buffer_name: str) -> None:
        """ Clears the given buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        raise NotImplementedError

    @dynamic_interface
    def create_buffer(self, buffer_name: str, size: int) -> None:
        """ Creates a reading buffer with the given name and size.

        Args:
            buffer_name (str): Name of the buffer
            size (size): Maximum number of readings (size >= 10) [default: 1000]
        """

        raise NotImplementedError

    @dynamic_interface
    def configure_sensors(self, channel_list: str, *, sense: Dict[str, List[Tuple]]):
        """ Configures the DAQ6510 to sense the specified channels.

        Args:
            channel_list (str): List of channels, as understood by the device
            sense (Dict[str], List[Tuple]): Dictionary with all the information on the configuration of the channels
        """

        raise NotImplementedError

    @dynamic_interface
    def setup_measurements(self, *, buffer_name: str, channel_list: str):
        """ Sets up the measurements for the given channel list.

        Args:
            buffer_name (str): Name of the buffer to use [default: defbuffer1]
            channel_list (str): Channels to read out
        """

        raise NotImplementedError

    @dynamic_interface
    def perform_measurement(self, *, buffer_name: str, channel_list: str, count: int, interval: int) -> list:
        """ Performs the actual measurements.

        Args:
            buffer_name (str): Name of the buffer
            channel_list (str): List of channels, as understood by the device
            count (int): Number of measurements to perform
            interval (int): Interval between measurements [s]
        """

        raise NotImplementedError


class DAQ6510Controller(DAQ6510Interface, DynamicCommandMixin):
    """
    The DAQ6510 Controller allows to remotely control the Keithley Data Acquisition System
    through an Ethernet interface.
    """

    def __init__(self, hostname: str = FW_SETTINGS.HOSTNAME, port: int = FW_SETTINGS.PORT):
        """ Opens a TCP/IP socket connection with the Keithley DAQ6510 Hardware.

        Args:
            hostname (str): IP address or fully qualified hostname of the Hexapod hardware controller. The default is
                            defined in the ``settings.yaml`` configuration file.
            port (int): IP port number to connect to, by default set in the ``settings.yaml`` configuration file.

        Raises:
            Error: when the connection could not be established for some reason.
        """

        super().__init__()

        logger.debug(f"Initializing the DAQ6510 Controller with hostname={hostname} on port={port}")

        self.daq = self.transport = DAQ6510EthernetInterface(hostname, port)

        # We set the default buffer here, this can be changed with the `create_buffer()` method.

        self.buffer_name = DEFAULT_BUFFER_1

    def is_simulator(self) -> bool:
        """ Indicates that the device is a real hardware controller

        Returns: False.
        """

        return False

    def connect(self) -> None:
        """ Connects to the device controller."""

        self.daq.connect()
        self.notify_observers(DeviceConnectionState.DEVICE_CONNECTED)

    def disconnect(self) -> None:
        """ Disconnects from the device controller."""

        self.daq.disconnect()
        self.notify_observers(DeviceConnectionState.DEVICE_NOT_CONNECTED)

    def reconnect(self) -> None:
        """ Reconnects to the device controller."""

        if self.is_connected():
            self.disconnect()

        self.connect()

    def is_connected(self) -> bool:
        """ Checks whether the device controller is connected.

        Returns: True if the device controller is connected; False otherwise.
        """

        return self.daq.is_connected()

    def send_command(self, command: str, response: bool) -> Union[None, str]:
        """ Sends an SCPI command to the device.

        The valid commands are described in the DAQ6510 Reference Manual [DAQ6510-901-01 Rev. B / September 2019].

        Args:
            command (str): SCPI command as specified in the DAQ6510 Reference Manual
            response (bool): Indicates whether you expect a reply from the device

        Returns: The response from the DAQ6510 is returned when a response was expected. When `response` is False,
                 None will be returned.
        """

        return self.daq.trans(command) if response else self.daq.write(command)

    def read_buffer(self, start: int, end: int, buffer_name: str = DEFAULT_BUFFER_1, elements: List[str] = None):
        """ Reads specific data elements (measurements) from the given buffer.

        Elements that can be specified to read out:

            - CHANNEL: Channel for which the data was acquired
            - DATE: Date when the data point was measured
            - READING: Actual reading of the measurement
            - TSTAMP: Timestamp when the data point was measured
            - UNIT: Unit of measure for the measurement
            - STATUS: Status information associated with the measurement

        Args:
            start: (int) First index of the buffer that should be returned (>= 1)
            end (int): Last index of the buffer that should be returned
            buffer_name (str): Name of the buffer to read out
            elements (List[str]): List of elements from the buffer to include in the response

        Returns: List of all the readings.
        """

        if elements is None:
            elements = ["READING"]
        else:
            elements = ", ".join(elements)

        return self.daq.trans(f'TRACE:DATA? {start}, {end}, "{buffer_name}", {elements}')

    def clear_buffer(self, buffer_name: str = DEFAULT_BUFFER_1) -> None:
        """ Clears the given buffer.

        Args:
            buffer_name (str): Name of the buffer
        """

        response = self.daq.trans(f'TRACE:ACTUAL? "{buffer_name}"')

        logger.info(f"Clearing buffer '{buffer_name}' containing {response} readings.")

        self.daq.write(f'TRACE:CLEAR "{buffer_name}"')

    def create_buffer(self, buffer_name: str, size: int = 1000) -> None:
        """ Creates a reading buffer with the given name.

        The name of the buffer must adhere to the following rules:

            - A buffer with this name should not exist in the device yet. When the buffer does exist, the DAQ6510 will
              show  a dialogue on the front panel with error 1115 saying the command cannot take an existing buffer
              name.
            - Buffer names must start with an alphabetic character.
            - The names cannot contain any periods nor the underscore (_) character.
            - The name can be up to 31 characters long.

        If the given size is 0, the instrument creates the largest reading buffer possible based on the available
        memory when the buffer is created.

        Args:
            buffer_name (str): Name of the buffer
            size (size): Maximum number of readings (size >= 10) [default: 1000]
        """

        self.daq.write(f'TRACE:MAKE "{buffer_name}", {size}')

        self.buffer_name = buffer_name

    # def reset(self) -> None:
    #
    #     self.daq.write("SYSTem:BEEPer 500, 0.1; :*RST; :SYSTem:BEEPer 1000, 0.1\n")

    def configure_sensors(self, channel_list: str, *, sense: Dict[str, List[Tuple]]) -> None:
        """ Configures the different sensors in the `channel_list`.

        Each sensor in the list will be configured according to the settings given in the `sense` dictionary.

        The following code will configure channels 101 and 102 as 4-wire transducers of type PT100.

        ```
        channel_list = create_channel_list(101, 102)

        sense = {
            "TEMPERATURE": [
                ("TRANSDUCER", "FRTD"),
                ("RTD:FOUR", "PT100"),
            ]
        }

        daq.configure_sensors(channel_list, sense=sense)
        ```

        The `sense` argument is a dictionary where the keys are function names like "TEMPERATURE" or "VOLTAGE:DC",
        and the values are a list of settings for that function. The list of settings is a list of tuples with the
        command and the value, e.g. "TRANSDUCER" is the settings command and "FRTD" is its value. The list of settings
        will be sent to the device in the order that they take in the list.

        Args:
            channel_list (str): Channels to configure
            sense: Dictionary with all the information on the configuration
        """

        if "TEMPERATURE" in sense:

            # Allowed settings for TEMPERATURE:
            #
            # - TEMPERATURE:APERTURE                  (@<channelList>)
            # - TEMPERATURE:AVERAGE:COUNT             (@<channelList>)
            # - TEMPERATURE:AVERAGE:STATE             (@<channelList>)
            # - TEMPERATURE:AVERAGE:TCONTROL          (@<channelList>)
            # - TEMPERATURE:AVERAGE:WINDOW            (@<channelList>)
            # - TEMPERATURE:AVERAGE:AZERO:STATE
            # - TEMPERATURE:DELAY:AUTO
            # - TEMPERATURE:DELAY:USER<N>
            # - TEMPERATURE:LINE:SYNC
            # - TEMPERATURE:NPLCYCLES
            # - TEMPERATURE:OCOMPENSATED
            # - TEMPERATURE:ODETECTOR
            # - TEMPERATURE:RELATIVE
            # - TEMPERATURE:RELATIVE:ACQUIRE
            # - TEMPERATURE:RELATIVE:STATE
            # - TEMPERATURE:RTD:ALPHA, BETA, DELTA      (@<channelList>)
            # - TEMPERATURE:RTD:ZERO, TWO, THREE, FOUR  (@<channelList>)
            # - TEMPERATURE:TCOUPLE:RJUNCTION:SIMULATED
            # - TEMPERATURE:TCOUPLE:RJUNCTION:RSELECT
            # - TEMPERATURE:TCOUPLE:TYPE
            # - TEMPERATURE:THERMISTOR      (@<channelList>)
            # - TEMPERATURE:TRANSDUCER      (@<channelList>)
            # - TEMPERATURE:UNIT            (@<channelList>)
            #

            # set the function to temperature

            self.daq.write(f'SENSE:FUNCTION "TEMPERATURE", {channel_list}')

            for cmd, value in sense["TEMPERATURE"]:
                self.daq.write(f"SENSE:TEMPERATURE:{cmd} {value}, {channel_list}")

    def setup_measurements(self, *, buffer_name: str = DEFAULT_BUFFER_1, channel_list: str) -> None:
        """ Sets up the measurements for the given channel list.

        Args:
            buffer_name (str): Name of the buffer to use [default: defbuffer1]
            channel_list (str): Channels to read out
        """

        self.daq.write(f'ROUTE:SCAN:BUFFER "{buffer_name}"')
        self.daq.write(f"ROUTE:SCAN:CREATE {channel_list}")
        self.daq.write(f"ROUTE:CHANNEL:OPEN {channel_list}")
        _ = self.daq.trans(f"ROUTE:CHANNEL:STATE? {channel_list}")
        self.daq.write("ROUTE:SCAN:START:STIMULUS NONE")

    def perform_measurement(self, *, buffer_name: str = DEFAULT_BUFFER_1, channel_list: str, count: int = 1,
                            interval: int = 2,) -> list:
        """ Performs the actual measurements.

        This function will wait until all measurements have completed, so be careful with the arguments `count` and
        `interval` as they will multiply into the number of seconds that you will have to wait for the response.

        Args:
            buffer_name (str): Name of the buffer
            channel_list (str): List of channels, as understood by the device
            count (int): Number of measurements to perform
            interval (int): Interval between measurements [s]

        Returns: List of readings.
        """

        # Set the number of times a scan is repeated

        self.daq.write(f"ROUTE:SCAN:COUNT:SCAN {count}")
        self.daq.write(f"ROUTE:SCAN:INTERVAL {interval}")  # [seconds]

        #
        self.daq.write("INITIATE:IMMEDIATE")
        self.daq.write("*WAI")

        # Read out the buffer

        logger.debug("Buffer count = ", self.get_buffer_count())

        num_sensors = count_number_of_channels(channel_list)

        readings = []

        for idx in range(1, count * num_sensors + 1):
            response = self.read_buffer(
                idx, idx, buffer_name=buffer_name, elements=["CHANNEL", "TSTAMP", "READING", "UNIT"]
            )
            if response != "" and response != str(count * num_sensors):
                if "\n" in response:
                    response = response.split("\n")
                    for i in range(len(response)):
                        readings.append(response[i].split(","))
                else:
                    readings.append(response.split(","))
                if len(readings[0]) < 4:
                    del readings[0]

        return readings


class DAQ6510Simulator(DAQ6510Interface):
    """
    Simulator for the Keithley DAQ6510 system.
    """

    def read_buffer(self, start: int, end: int, buffer_name: str, elements: List[str]):
        pass

    def get_buffer_count(self, buffer_name: str = DEFAULT_BUFFER_1):
        pass

    def get_buffer_capacity(self, buffer_name: str):
        pass

    def delete_buffer(self, buffer_name: str):
        pass

    def clear_buffer(self, buffer_name: str):
        pass

    def create_buffer(self, buffer_name: str, size: int):
        pass

    def configure_sensors(self, channel_list: str, *, sense: Dict[str, List[Tuple]]):
        pass

    def setup_measurements(self, *, buffer_name: str, channel_list: str):
        pass

    def perform_measurement(self, *, buffer_name: str, channel_list: str, count: int, interval: int):
        pass

    def send_command(self, command: str, response: bool):
        pass

    def info(self) -> str:
        pass

    def reset(self):
        pass

    def is_simulator(self):
        """ Indicates that the device is a simulator.

        Returns: True.
        """

        return True

    def connect(self):
        pass

    def disconnect(self):
        pass

    def reconnect(self):
        pass

    def is_connected(self):
        pass


class DAQ6510Proxy(Proxy, DAQ6510Interface):
    """
    The DAQ6510Proxy class is used to connect to the Keithley Control Server and send commands
    to the Keithley Hardware Controller remotely.
    """

    def __init__(
        self,
        protocol: str = CTRL_SETTINGS.PROTOCOL,
        hostname: str = CTRL_SETTINGS.HOSTNAME,
        port: int = CTRL_SETTINGS.COMMANDING_PORT,
        timeout: int = CTRL_SETTINGS.TIMEOUT * 1000    # Timeout [ms]: > scan count * interval + (one scan duration)
    ):
        """ Initialisation of a DAQ6510Proxy.

        Args:
            protocol (str): Transport protocol [default is taken from settings file]
            hostname (str): Location of the Control Server (IP address) [default is taken from settings file]
            port (int): TCP port on which the Control Server is listening for commands [default is taken from settings
                        file]
            timeout (int): Timeout by which to establish the connection [ms]
        """

        super().__init__(connect_address(protocol, hostname, port), timeout=timeout)


def create_channel_list(*args) -> str:
    """ Createa a channel list that is understood by the SCPI commands of the DAQ6510.

    Channel names contain both the slot number and the channel number. The slot number is the number of the slot where
    the card is installed at the back of the device.

    When addressing multiple individual channels, add each of them as a separate argument, e.g. to include channels 1,
    3, and 7 from slot 1, use the following command:

        >>> create_channel_list(101, 103, 107)
        '(@101, 103, 107)'

    To designate a range of channels, only one argument should be given, i.e. a tuple containing two channels
    representing the range. The following tuple `(101, 110)` will create the following response: `"(@101:110)"`. The
    range is inclusive, so this will define a range of 10 channels in slot 1.

        >>> create_channel_list((201, 205))
        '(@201:205)'

    See reference manual for the Keithley DAQ6510 [DAQ6510-901-01 Rev. B / September 2019], chapter 11: Introduction to
    SCPI commands, SCPI command formatting, channel naming.

    Args:
        *args: Tuple or a list of channels

    Returns: String containing the channel list as understood by the device.
    """

    if not args:
        return ""

    # If only one argument is given, I expect either a tuple defining a range or just one channel. When several
    # arguments are given, I expect them all to be individual channels.

    if len(args) == 1:

        arg = args[0]
        if isinstance(arg, tuple):
            ch_list = f"(@{arg[0]}:{arg[1]})"
        else:
            ch_list = f"(@{arg})"

    else:
        ch_list = "(@" + ", ".join([str(arg) for arg in args]) + ")"

    return ch_list


def count_number_of_channels(channel_list: str) -> int:
    """ Given a proper channel list, this function counts the number of channels.

    For ranges, it returns the actual number of channels that are included in the range.

        >>> count_number_of_channels("(@1,2,3,4,5)")
        5
        >>> count_number_of_channels("(@1, 3, 5)")
        3
        >>> count_number_of_channels("(@2:7)")
        6

    Args:
        channel_list (str): Channel list as understood by the SCPI commands of DAQ6510

    Returns: Number of channels in the list.
    """

    match = re.match(r"\(@(.*)\)", channel_list)
    group = match.groups()[0]

    parts = group.replace(" ", "").split(",")
    count = 0
    for part in parts:
        if ":" in part:
            split_part = part.split(":")
            count += int(split_part[1]) - int(split_part[0]) + 1
        else:
            count += 1

    return count


def get_channel_names(channel_list: str) -> List[str]:
    """ Generates a list of channel names from a given channel list.

    Args:
        channel_list (str): Channel list as understood by the SCPI commands of DAQ6510

    Returns: List of channel names.
    """

    match = re.match(r"\(@(.*)\)", channel_list)
    group = match.groups()[0]

    parts = group.replace(" ", "").split(",")
    names = []
    for part in parts:
        if ":" in part:
            split_part = part.split(":")
            names.extend(str(ch) for ch in range(int(split_part[0]), int(split_part[1]) + 1))
        else:
            names.append(part)

    return names


if __name__ == "__main__":

    logging.basicConfig(level=20)

    print(f'{get_channel_names("(@101:105)")=}')
    print(f'{get_channel_names("(@101, 102, 103, 105)")=}')
    # sys.exit(0)

    daq = DAQ6510Controller()
    daq.connect()
    daq.reset()

    print(daq.info())

    buffer_capacity = daq.get_buffer_capacity()
    print(f"buffer {DEFAULT_BUFFER_1} can still hold {buffer_capacity} readings")

    buffer_count = daq.get_buffer_count()
    print(f"buffer {DEFAULT_BUFFER_1} holds {buffer_count} readings")

    channels = create_channel_list((101, 102))

    print(channels)

    sense_dict = {"TEMPERATURE": [("TRANSDUCER", "FRTD"), ("RTD:FOUR", "PT100"), ("UNIT", "KELVIN")]}

    daq.configure_sensors(channels, sense=sense_dict)

    daq.setup_measurements(channel_list=channels)

    meas_response = daq.perform_measurement(channel_list=channels, count=5, interval=1)

    print(meas_response)

    buffer_count = daq.get_buffer_count()
    print(f"buffer {DEFAULT_BUFFER_1} holds {buffer_count} readings")

    daq.disconnect()



# Test if object can be pickled.
import pickle

from egse.response import Failure
from egse.response import Message
from egse.response import Success


def test_failure():

    try:
        1/0
    except ZeroDivisionError as exc:
        failure = Failure("You cannot divide one by zero")

        assert f"{failure!r}" == "Failure('You cannot divide one by zero')"
        assert failure.message == str(failure) == f"{failure}" == "You cannot divide one by zero"
        assert failure.cause is None
        assert not failure.successful

        data = pickle.dumps(failure)

        assert isinstance(pickle.loads(data), Failure)

        failure = pickle.loads(data)

        assert f"{failure!r}" == "Failure('You cannot divide one by zero')"
        assert failure.message == str(failure) == f"{failure}" == "You cannot divide one by zero"
        assert failure.cause is None
        assert not failure.successful

        # ----- adding the cause

        failure = Failure("You cannot divide one by zero", cause=exc)

        assert f"{failure!r}" == ("Failure('You cannot divide one by zero: division by zero', "
                                  "cause=ZeroDivisionError('division by zero'))")
        assert failure.message == str(failure) == f"{failure}" == "You cannot divide one by zero: division by zero"
        assert isinstance(failure.cause, ZeroDivisionError)
        assert f"{failure.cause!r}" == "ZeroDivisionError('division by zero')"
        assert str(failure.cause) == f"{failure.cause!s}" == "division by zero"
        assert not failure.successful

        data = pickle.dumps(failure)

        assert isinstance(pickle.loads(data), Failure)

        failure = pickle.loads(data)

        assert f"{failure!r}" == ("Failure('You cannot divide one by zero: division by zero', "
                                  "cause=ZeroDivisionError('division by zero'))")
        assert failure.message == str(failure) == f"{failure}" == "You cannot divide one by zero: division by zero"
        assert isinstance(failure.cause, ZeroDivisionError)
        assert f"{failure.cause!r}" == "ZeroDivisionError('division by zero')"
        assert str(failure.cause) == f"{failure.cause!s}" == "division by zero"
        assert not failure.successful


def test_success():

    success = Success("This is a great success!")

    assert f"{success!r}" == "Success('This is a great success!')"
    assert success.message == str(success) == f"{success!s}" == "This is a great success!"
    assert success.response is None
    assert success.return_code is None
    assert success.successful

    data = pickle.dumps(success)

    assert isinstance(pickle.loads(data), Success)

    success = pickle.loads(data)

    assert f"{success!r}" == "Success('This is a great success!')"
    assert success.message == str(success) == f"{success!s}" == "This is a great success!"
    assert success.response is None
    assert success.return_code is None
    assert success.successful

    # ----- adding a return_code / response

    success = Success("This is a great success", return_code=3.1415)

    assert f"{success!r}" == "Success('This is a great success: 3.1415', return_code=3.1415)"
    assert success.message == str(success) == f"{success!s}" == "This is a great success: 3.1415"
    assert success.response == 3.1415
    assert success.return_code == 3.1415
    assert success.successful

    data = pickle.dumps(success)

    assert isinstance(pickle.loads(data), Success)

    success = pickle.loads(data)

    assert f"{success!r}" == "Success('This is a great success: 3.1415', return_code=3.1415)"
    assert success.message == str(success) == f"{success!s}" == "This is a great success: 3.1415"
    assert success.response == 3.1415
    assert success.return_code == 3.1415
    assert success.successful


def test_message():

    msg = Message("Hello, World!")

    assert f"{msg!r}" == "Message('Hello, World!')"
    assert str(msg) == f"{msg}" == "Hello, World!"



"""
This module provides functionality to handle responses from the control servers.
"""
__all__ = [
    "Failure",
    "Message",
    "Response",
    "Success",
]

from typing import Any

import rich.repr


class Response:
    """
    Base class for any reply or response between client-server communication.

    The idea is that the response is encapsulated in one of the subclasses depending
    on the type of response.
    """

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return self.message

    @property
    def successful(self):
        """Returns True if the Response is not an Exception."""
        return not isinstance(self, Exception)


@rich.repr.auto
class Failure(Response, Exception):
    """
    A failure response indicating something went wrong at the other side.

    This class is used to encapsulate an Exception that was caught and needs to be
    passed to the client. So, the intended use is like this:

        try:
            # perform some useful action that might raise an Exception
        except SomeException as exc:
            return Failure("Our action failed", exc)

    The client can inspect the Exception that was originally raised, in this case `SomeException`
    with the `cause` variable.

    Since a Failure is also an Exception, the property `successful` will return False.
    So, the calling method can test for this easily.

        rc: Response = function_that_returns_a_response()

        if not rc.successful:
            # handle the failure
        else:
            # handle success

    """

    def __init__(self, message: str, cause: Exception = None):
        msg = f"{message}: {cause}" if cause is not None else message
        super().__init__(msg)
        self.cause = cause


@rich.repr.auto
class Success(Response):
    """
    A success response for the client.

    The return code from any action or function that needs to be returned to the
    client shall be added.

    Since `Success` doesn't inherit from `Exception`, the property `successful` will return True.
    """

    def __init__(self, message: str, return_code: Any = None):
        msg = f"{message}: {return_code}" if return_code is not None else message
        super().__init__(msg)
        self.return_code = return_code

    @property
    def response(self):
        return self.return_code


@rich.repr.auto
class Message(Response):
    """
    A message response from the client.

    Send a Message when there is no Failure, but also no return code. This is the alternative of
    returning a None.

    Message returns True for the property successful since it doesn't inherit from Exception.
    """

    pass



import json
import sys
import time
import traceback


def main():
    time.sleep(1.0)

    try:
        raise ValueError("for testing purposes")
    except Exception as e:
        error_info = {
            "type": type(e).__name__,
            "message": str(e),
            "traceback": traceback.format_exc()
        }
        print(json.dumps(error_info), file=sys.stderr)
        sys.exit(1)


if __name__ == '__main__':
    main()



import logging

from prometheus_client import start_http_server

import multiprocessing
multiprocessing.current_process().name = "daq6510_cs"

import sys

import click
import invoke
import rich
import zmq

from egse.control import ControlServer
from egse.control import is_control_server_active
from egse.settings import Settings
from egse.tempcontrol.keithley.daq6510 import DAQ6510Proxy
from egse.tempcontrol.keithley.daq6510_protocol import DAQ6510Protocol
from egse.zmq_ser import connect_address

logger = logging.getLogger(__name__)

CTRL_SETTINGS = Settings.load("Keithley Control Server")


def is_daq6510_cs_active(timeout: float = 0.5) -> bool:
    """ Checks if the DAQ6510 Control Server is running.

    Args:
        timeout (float): Timeout when waiting for a reply [s, default=0.5]

    Returns:  True if the Control Server is running and replied with the expected answer; False otherwise.
    """

    endpoint = connect_address(
        CTRL_SETTINGS.PROTOCOL, CTRL_SETTINGS.HOSTNAME, CTRL_SETTINGS.COMMANDING_PORT
    )

    return is_control_server_active(endpoint, timeout)


class DAQ6510ControlServer(ControlServer):
    """
    Keithley DAQ6510ControlServer - Command and monitor the Keithley Data Acquisition System.

    This class works as a command and monitoring server to control the DAQ6510 Controller.

    The sever binds to the following ZeroMQ sockets:

        - REQ-REP socket that can be used as a command server. Any client can connect and send a command to the
          DAQ6510 controller.

        - PUB-SUP socket that serves as a monitoring server. It will send out DAQ6510 status information to all the
          connected clients every DELAY seconds.

    """

    def __init__(self):
        """ Initialisation of a DAQ6510 Control Server."""

        super().__init__()

        self.device_protocol = DAQ6510Protocol(self)

        self.logger.info(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

    def get_communication_protocol(self) -> str:
        """ Returns the communication protocol used by the Control Server.

        Returns: Communication protocol used by the Control Server, as specified in the settings.
        """

        return CTRL_SETTINGS.PROTOCOL

    def get_commanding_port(self) -> int:
        """ Returns the commanding port used by the Control Server.

        Returns: Commanding port used by the Control Server, as specified in the settings.
        """

        return CTRL_SETTINGS.COMMANDING_PORT

    def get_service_port(self):
        """ Returns the service port used by the Control Server.

        Returns: Service port used by the Control Server, as specified in the settings.
        """

        return CTRL_SETTINGS.SERVICE_PORT

    def get_monitoring_port(self):
        """ Returns the monitoring port used by the Control Server.

        Returns: Monitoring port used by the Control Server, as specified in the settings.
        """

        return CTRL_SETTINGS.MONITORING_PORT

    def get_storage_mnemonic(self):
        """ Returns the storage mnemonics used by the Control Server.

        This is a string that will appear in the filename with the housekeeping information of the device, as a way of
        identifying the device.  If this is not implemented in the sub-class, then the class name will be used.

        Returns: Storage mnemonics used by the Control Server, as specified in the settings.  If not specified in the
                 settings, "DAQ6510" will be used.
        """

        try:
            return CTRL_SETTINGS.STORAGE_MNEMONIC
        except AttributeError:
            return "DAQ6510"

    def before_serve(self):
        """ Steps to take before the Control Server is activated."""

        start_http_server(CTRL_SETTINGS.METRICS_PORT)


@click.group()
def cli():
    pass


@cli.command()
def start():
    """ Starts the Keithley DAQ6510 Control Server."""

    try:
        control_server = DAQ6510ControlServer()
        control_server.serve()
    except KeyboardInterrupt:
        logger.debug("Shutdown requested...exiting")
    except SystemExit as exit_code:
        logger.debug("System Exit with code {}.".format(exit_code))
        sys.exit(exit_code)
    except Exception:
        msg = "Cannot start the DAQ6510 Control Server"
        logger.exception(msg)
        rich.print(f"[red]{msg}.")

    return 0


@cli.command()
def start_bg():
    """ Starts the DAQ6510 Control Server in the background."""

    invoke.run("daq6510_cs start", disown=True)


@cli.command()
def stop():
    """ Sends a 'quit_server' command to the Keithley DAQ6510 Control Server."""

    try:
        with DAQ6510Proxy() as daq:
            sp = daq.get_service_proxy()
            sp.quit_server()
    except ConnectionError as exc:
        msg = "Cannot stop the DAQ6510 Control Server"
        logger.exception(msg)
        rich.print(f"[red]{msg}, could not send the Quit command. [black]Check log messages.")


@cli.command()
def status():
    """ Requests status information from the Control Server."""

    protocol = CTRL_SETTINGS.PROTOCOL
    hostname = CTRL_SETTINGS.HOSTNAME
    port = CTRL_SETTINGS.COMMANDING_PORT

    endpoint = connect_address(protocol, hostname, port)

    if is_control_server_active(endpoint):
        rich.print(f"DAQ6510 CS: [green]active")
        with DAQ6510Proxy() as daq6510:
            sim = daq6510.is_simulator()
            connected = daq6510.is_connected()
            ip = daq6510.get_ip_address()
            rich.print(f"mode: {'simulator' if sim else 'device'}{' not' if not connected else ''} connected")
            rich.print(f"hostname: {ip}")
    else:
        rich.print(f"DAQ6510 CS: [red]not active")


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG, format=Settings.LOG_FORMAT_FULL)

    sys.exit(cli())



"""
The referenceFrames module provides the class :code:`ReferenceFrames` which defines the affine transformation
for bringing one reference frame to another.

.. todo:: The tests in methods like getPassiveTransformationTo using '==' should be looked at again and maybe
          changed into using the 'is' operator. This because we now have __eq__ implemented.

@author: Pierre Royer
"""
import logging
import random
import string
import textwrap

import numpy as np
import transforms3d as t3

import egse.coordinates.transform3d_addon as t3add
from egse.coordinates.point import Point
from egse.coordinates.rotationMatrix import RotationMatrix
from egse.decorators import deprecate
from egse.exceptions import InvalidOperationError

LOGGER = logging.getLogger(__name__)
DEBUG = False


def transformationToString(transformation):
    """Helper function: prints out a transformation (numpy ndarray) in a condensed form on one line."""

    if isinstance(transformation, np.ndarray):
        if np.allclose(transformation, ReferenceFrame._I):
            return "Identity"
        msg = np.array2string(
            transformation,
            separator=",",
            suppress_small=True,
            formatter={"float_kind": lambda x: "%.2f" % x},
        ).replace("\n", "")
        return msg

    # We do not want to raise an Exception here since this is mainly used in logging messages
    # and doesn't really harm the execution of the program.
    return f"ERROR: expected transformation to be an ndarray, type={type(transformation)}"


class ReferenceFrame(object):
    """
    A Reference Frame defined in reference frame "ref", i.e.
    defined by the affine transformation bringing the reference frame "ref" onto "self".

    By default, "ref" is the master refence frame, defined as the identity matrix.

    :param transformation: 4x4 affine transformation matrix defining this system in "ref" system
    :type transformation: numpy array

    :param ref: reference system in which this new reference frame is defined
    :type ref: ReferenceFrame

    :param name: name the reference frame so it can be referenced, set to 'master' when None
    :type name: str

    :param rot_config:
            * Is set when using creator ReferenceFrame.fromTranslationRotation()
            * In other cases, is set to a default "szyx"
              (rotations around static axes z, y and x in this order)
              In these other cases, it has no real direct influence,
              except for methods returning the rotation vector (e.g. getRotationVector)
              It is therefore always recommended to pass it to the constructor, even when
              constructing the ReferenceFrame directly from a transformation matrix
    :type rot_config: str

    Both the ``transformation`` and the ``ref`` parameters are mandatory.

    If the reference frame is None, the master reference frame is created.

    The master reference frame:

        * is defined by the identity transformation matrix
        * has itself as a reference

    For convenience we provide the following factory methods:

    createMaster()
        Create a Master Reference Frame

    createRotation(..)
        Create a new Reference Frame that is rotated with respect to the given reference frame

    createTranslation(..)
        Create a new Reference Frame that is a translation with respect to the given reference frame
    """

    _I = np.identity(4)
    _MASTER = None
    _ROT_CONFIG_DEFAULT = "sxyz"
    _names_used = [None, "Master"]
    _strict_naming = False
    _ACTIVE_DEFAULT = True

    def __new__(cls, transformation, ref, name=None, rot_config=_ROT_CONFIG_DEFAULT):
        """Create a new ReferenceFrame class."""

        DEBUG and LOGGER.debug(
            f"transformation={transformationToString(transformation)}, ref={ref!r}, name={name}, rot_config={rot_config}"
        )

        if ref is None:
            msg = (
                "No reference frame was given, if you planned to create a Master Reference Frame, "
                "use ReferenceFrame.createMaster(). "
            )
            LOGGER.error(msg)
            raise ValueError(msg, "REF_IS_NONE")

        if not isinstance(ref, cls):
            msg = f"The 'ref' keyword argument is not a ReferenceFrame object, but {type(ref)}"
            LOGGER.error(msg)
            raise ValueError(msg, "REF_IS_NOT_CLS")

        if name == "Master":
            msg = (
                "The 'name' argument cannot be 'Master' unless a Master instance should be created, "
                "in that case, use ReferenceFrame.createMaster()"
            )
            LOGGER.error(msg)
            raise ValueError(msg, "MASTER_NAME_USED")

        if transformation is None:
            msg = "The 'transformation' argument can not be None, please provide a proper transformation for this reference frame."
            LOGGER.error(msg)
            raise ValueError(msg, "TRANSFORMATION_IS_NONE")

        if not isinstance(transformation, np.ndarray):
            msg = f"The 'transformation' argument shall be a Numpy ndarray [not a {type(transformation)}], please provide a proper transformation for this reference frame."
            LOGGER.error(msg)
            raise ValueError(msg, "TRANSFORMATION_IS_NOT_NDARRAY")

        if rot_config is None:
            msg = "The 'rot_config' keyword argument can not be None, do not specify it when you want to use the default value."
            LOGGER.error(msg)
            raise ValueError(msg)

        _instance = super(ReferenceFrame, cls).__new__(cls)

        return _instance

    def __init__(self, transformation, ref, name=None, rot_config=_ROT_CONFIG_DEFAULT):
        """Initialize the ReferenceFrame object """

        self.debug = False

        DEBUG and LOGGER.debug(
            f"transformation={transformationToString(transformation)}, ref={ref!r}, name={name}, rot_config={rot_config}"
        )

        # All argument testing is done in the __new__() method and we should be save here.

        self.ref = ref
        self.name = self.__createName(name)
        self.transformation = transformation
        self.rot_config = rot_config

        self.definition = [self.transformation, self.ref, self.name]

        self.x = self.getAxis("x")
        self.y = self.getAxis("y")
        self.z = self.getAxis("z")

        self.linkedTo = {}
        self.referenceFor = []

        ref.referenceFor.append(self)

        return

    def find_master(self):
        return self.findMaster()

    def findMaster(self):
        """
        Returns the Master frame for this reference frame. The Master frame is always at the end
        of the path following the references.

        Returns:
            The master frame.
        """

        frame = self
        while not frame.isMaster():
            frame = frame.ref
        return frame

    @classmethod
    def createMaster(cls):
        """
        Create a master reference frame.

        A master reference frame is defined with respect to itself and is initialised with the
        identity matrix.

        The master frame is automatically given the name "Master".
        """
        ref_master = super(ReferenceFrame, cls).__new__(cls)
        ref_master.name = "Master"
        ref_master.ref = ref_master
        ref_master.transformation = cls._I
        ref_master.rot_config = cls._ROT_CONFIG_DEFAULT
        ref_master.initialized = True
        ref_master.debug = False
        ref_master.linkedTo = {}
        ref_master.referenceFor = []

        DEBUG and LOGGER.debug(
            f"NEW MASTER CREATED: {id(ref_master)}, ref = {id(ref_master.ref)}, "
            f"name = {ref_master.name}"
        )

        return ref_master

    @classmethod
    def __createName(cls, name: str = None):

        if name is None:
            while name in cls._names_used:
                name = "F" + "".join(random.choices(string.ascii_uppercase, k=3))
            return name

        if cls._strict_naming:

            # generate a unique name

            old_name = name

            while name in cls._names_used:
                name = "F" + "".join(random.choices(string.ascii_uppercase, k=3))

            LOGGER.warning(
                f"name ('{old_name}') is already defined, since strict naming is applied, "
                f"a new unique name was created: {name}"
            )

        else:

            if name in cls._names_used:
                DEBUG and LOGGER.warning(
                    f"name ('{name}') is already defined, now you have more than one "
                    f"ReferenceFrame with the same name.",
                )

        cls._names_used.append(name)
        return name

    @classmethod
    def fromTranslation(cls, transx, transy, transz, ref, name=None):
        """
        Create a ReferenceFrame from a translation with respect to the given reference frame.

        :param transx: translation along the x-axis
        :type transx: float

        :param transy: translation along the y-axis
        :type transy: float

        :param transz: translation along the z-axis
        :type transz: float

        :param ref: reference frame with respect to which the translation is performed. If no ref is given
            the rotation is with respect to the master reference frame
        :type ref: ReferenceFrame

        :param name: a simple convenient name to identify the reference frame. If no name is provided
                     a random name of four characters starting with 'F' will be generated.
        :type name: str

        :return: a reference frame
        """
        adef = np.identity(4)
        adef[:3, 3] = [transx, transy, transz]

        if ref is None:
            raise ValueError(
                "The ref argument can not be None, provide a master or another reference frame."
            )

        return cls(transformation=adef, ref=ref, name=name)

    @classmethod
    def fromRotation(
        cls,
        rotx,
        roty,
        rotz,
        ref,
        name=None,
        rot_config=_ROT_CONFIG_DEFAULT,
        active=_ACTIVE_DEFAULT,
        degrees=True,
    ):
        """
        Create a ReferenceFrame from a rotation with respect to the given reference frame.

        :param rotx: rotation around the x-axis
        :type rotx: float

        :param roty: rotation around the y-axis
        :type roty: float

        :param rotz: rotation around the z-axis
        :type rotz: float

        :param ref: reference frame with respect to which the rotation is performed. If no ref is given
            the rotation is with respect to the master reference frame
        :type ref: ReferenceFrame

        :param name: a simple convenient name to identify the reference frame. If no name is provided
                     a random name of four characters starting with 'F' will be generated.
        :type name: str

        :return: a reference frame
        """
        # Convention (rotating axes, in order xyz)

        # rot_config = cls._ROT_CONFIG_DEFAULT

        # Rotation amplitude

        # Rotation
        if degrees:
            rotx = np.deg2rad(rotx)
            roty = np.deg2rad(roty)
            rotz = np.deg2rad(rotz)

        rotation = RotationMatrix(rotx, roty, rotz, rot_config, active=active)

        # Defaults for zoom & shear

        zdef = np.array([1, 1, 1])
        sdef = np.array([0, 0, 0])

        translation = [0, 0, 0]

        TT = t3.affines.compose(T=translation, R=rotation.R, Z=zdef, S=sdef)

        if ref is None:
            raise ValueError(
                "The ref argument can not be None, provide a master or another reference frame."
            )

        return cls(transformation=TT, ref=ref, name=name, rot_config=rot_config)

    @staticmethod
    def fromPoints(points, plane="xy", name=None, usesvd=True, verbose=True):
        """
        fromPoints(points, plane="xy", name=None, usesvd=True,verbose=True)

        Finds the best fitting "plane" plane to all points in "points" and builds a ReferenceFrame
        considering the fitted plane as the 'xy' plane of the new system

        fitPlane must be in ['xy','yz','zx']
        """
        return points.bestFittingPlane(fitPlane=plane, usesvd=usesvd, verbose=verbose)

    @classmethod
    def fromTranslationRotation(
        cls,
        translation,
        rotation,
        ref,
        name=None,
        rot_config=_ROT_CONFIG_DEFAULT,
        active=_ACTIVE_DEFAULT,
        degrees=True,
    ):
        """fromTranslationRotation(cls,translation,rotation, ref=None,name=None, rot_config=_ROT_CONFIG_DEFAULT,active=_ACTIVE_DEFAULT, degrees=True)

        Construct a ReferenceFrame from a translation and a rotation (vectors!)

        translation : translation vector : 3 x 1 = tx, ty, tz
        rotation    : rotation vector    : 3 x 1 = rx, ry, rz

        rot_config : convention for the rotation, see rotationMatrix.__doc__

        degrees : if True, input values are assumed in degrees, otherwise radians
        """
        # Translation
        translation = np.array(translation)
        # Zoom - unit
        zdef = np.array([1, 1, 1])
        # Shear
        sdef = np.array([0, 0, 0])
        # Rotation
        if degrees:
            rotation = np.array([np.deg2rad(item) for item in rotation])
        rotx, roty, rotz = rotation
        #
        rmat = RotationMatrix(rotx, roty, rotz, rot_config=rot_config, active=active)
        #
        # transformation = t3.affines.compose(translation,rmat.R,Z=zdef,S=sdef)

        if ref is None:
            raise ValueError(
                "The ref argument can not be None, provide a master or another reference frame."
            )

        return cls(
            transformation=t3.affines.compose(translation, rmat.R, Z=zdef, S=sdef),
            ref=ref,
            name=name,
            rot_config=rot_config,
        )

    def getTranslationVector(self):
        """
        getTranslationVector()

        Returns the translation defining this reference frame (from self.ref to self)
        """
        return self.transformation[:3, 3]

    def getRotationVector(self, degrees=True, active=_ACTIVE_DEFAULT):
        """
        getRotationVector()

        Returns the rotation (vector) defining this reference frame (from self.ref to self)
        degrees : if true, rotation angles are expressed in degrees
        """
        rotation = t3.euler.mat2euler(self.transformation, axes=self.rot_config)
        if degrees:
            rotation = np.array([np.rad2deg(item) for item in rotation])
        return rotation

    def getTranslationRotationVectors(self, degrees=True, active=_ACTIVE_DEFAULT):
        """getTranslationRotationVectors()
        Returns the translation and rotation (vectors) defining this reference frame (from self.ref to self)
        degrees : if true, rotation angles are expressed in degrees
        """
        translation = self.getTranslationVector()
        rotation = self.getRotationVector(degrees=degrees, active=active)
        return translation, rotation

    def getRotationMatrix(self):
        """
        getRotationMatrix()

        Returns the rotation (matrix) defining this reference frame (from self.ref to self)
        """
        result = self.transformation.copy()
        result[:3, 3] = [0.0, 0.0, 0.0]
        return result

    def __repr__(self):
        return f"ReferenceFrame(transformation={transformationToString(self.transformation)}, ref={self.ref.name}, name={self.name}, rot_config={self.rot_config})"

    def __str__(self):
        msg = textwrap.dedent(
            f"""\
                ReferenceFrame
                name          : {self.name}
                reference     : {self.ref.name}
                rot_config    : {self.rot_config}
                links         : {[key.name for key in self.linkedTo.keys()]}
                transformation:
                  [{np.round(self.transformation[0], 3)}
                   {np.round(self.transformation[1], 3)}
                   {np.round(self.transformation[2], 3)}
                   {np.round(self.transformation[3], 3)}]
                translation   : {np.round(self.getTranslationVector(),3)}
                rotation      : {np.round(self.getRotationVector(),3)}"""
        )
        return msg

    @deprecate(
        reason=(
            "I do not see the added value of changing the name and "
            "the current method has the side effect to change the name "
            "to a random string when the name argument is already used."
        ),
        alternative="the constructor argument to set the name already of the object.",
    )
    def setName(self, name=None):
        """
        Set or change the name of a reference frame.

        ..note: this method is deprecated

        :param str name: the new name for the reference frame, if None, a random name will be generated.

        :raises InvalidOperationError: when you try to change the name of the Master reference frame
        """
        if self.isMaster():
            raise InvalidOperationError(
                "You try to change the name of the Master reference frame, which is not allowed."
            )
        self.name = self.__createName(name)

    def addLink(self, ref, transformation=None, _stop=False):
        """
        adds a link between self and ref in self.linkedTo and ref.linkedTo
        """

        # DONE: set the inverse transformation in the ref to this
        #   ref.linkedTo[self] = t3add.affine_inverse(transformation)
        # TODO:
        #   remove the _stop keyword

        # TODO: deprecate transformation as an input variable
        #       linkedTo can become a list of reference frames, with no transformation
        #       associated to the link. The tfo associated to a link is already
        #       checked in real time whenever the link is addressed
        if transformation is None:
            transformation = self.getActiveTransformationTo(ref)
        else:
            if DEBUG:
                LOGGER.info("Deprecation warning: transformation will be automatically set to "
                            "the current relation between {self.name} and {ref.name}")
                LOGGER.debug("Requested:")
                LOGGER.debug(np.round(transformation, decimals=3))
                LOGGER.debug("Auto (enforced):")

            transformation = self.getActiveTransformationTo(ref)

            DEBUG and LOGGER.debug(np.round(transformation, decimals=3))

        self.linkedTo[ref] = transformation

        # TODO simplify this when transformation is deprecated
        #      it becomes ref.linkedTo[self] = ref.getActiveTransformationTo(self)
        ref.linkedTo[self] = t3add.affine_inverse(transformation)

    def removeLink(self, ref):
        """
        Remove the link between self and ref, both ways.
        """

        # First remove the entry in ref to this

        if self in ref.linkedTo:
            del ref.linkedTo[self]

        # Then remove the entry in this to ref

        if ref in self.linkedTo:
            del self.linkedTo[ref]

    def getPassiveTransformationTo(self, targetFrame):
        """
        getPassiveTransformationTo(self,targetFrame)
        == getPointTransformationTo(self,targetFrame)

        returns the transformation to apply to a Point (defined in self) to express it in targetFrame

        Passive transformation : the point is static, we change the reference frame around it
        """
        DEBUG and LOGGER.debug("PASSIVE TO self {self.name} target {targetFrame.name}")
        if targetFrame is self:
            """
            Nothing to do here, we already have the right coordinates
            """
            DEBUG and LOGGER.debug("case 1")
            result = np.identity(4)

        elif targetFrame.ref is self:
            """
            The target frame is defined in self => the requested transformation is the targetFrame definition
            """
            DEBUG and LOGGER.debug("=== 2 start ===")
            result = t3add.affine_inverse(targetFrame.transformation)
            DEBUG and LOGGER.debug("=== 2 end   ===")
        elif targetFrame.ref is self.ref:
            """
            targetFrame and self are defined wrt the same reference frame
            We want
            self --> targetFrame
            We know
            targetFrame.ref --> targetFrame (= targetFrame.transformation)
            self.ref   --> self   (= self.transformation)
            That is
            self --> self.ref is targetFrame.ref --> targetFrame
            inverse(definition)    targetFrame definition

            """
            if DEBUG:
                LOGGER.debug("=== 3 start ===")
                LOGGER.debug(" ref   \n{0}".format(self.ref))
                LOGGER.debug("===")
                LOGGER.debug("self   \n{0}".format(self))
                LOGGER.debug("===")
                LOGGER.debug("targetFrame \n{0}".format(targetFrame))
                LOGGER.debug("===")

            selfToRef = self.transformation
            DEBUG and LOGGER.debug("selfToRef \n{0}".format(selfToRef))

            # refToRef = I

            refToTarget = t3add.affine_inverse(targetFrame.transformation)
            DEBUG and LOGGER.debug("refToTarget \n{0}".format(refToTarget))

            result = np.dot(refToTarget, selfToRef)
            DEBUG and LOGGER.debug("result \n{0}".format(result))
            DEBUG and LOGGER.debug("=== 3 end   ===")
        else:
            """
            We are after the transformation from
            self --> targetFrame
            ==
            self --> self.ref --> targetFrame.ref --> targetFrame

            We know
            targetFrame.ref --> targetFrame (targetFrame.transformation)
            self.ref        --> self (self.transformation)
            but
            targetFrame.ref != self.ref
            so we need
            self.ref --> targetFrame.ref
            then we can compose
            self --> self.ref --> targetFrame.ref --> targetFrame

            Note: the transformation self.ref --> targetFrame.ref is acquired recursively
                  This relies on the underlying assumption that there exists
                  one unique reference frame that source and self can be linked to
                  (without constraints on the number of links necessary), i.e.
                  that, from a frame to its reference or the opposite, there exists
                  a path between self and targetFrame. That is equivalent to
                  the assumption that the entire set of reference frames is connex,
                  i.e. defined upon a unique master reference frame.
            """
            DEBUG and LOGGER.debug("=== 4 start ===")
            selfToRef = self.transformation
            selfRefToTargetRef = self.ref.getPassiveTransformationTo(targetFrame.ref)
            refToTarget = t3add.affine_inverse(targetFrame.transformation)
            result = np.dot(refToTarget, np.dot(selfRefToTargetRef, selfToRef))
            DEBUG and LOGGER.debug("=== 4 end   ===")

        return result

    def getPassiveTranslationRotationVectorsTo(
        self, targetFrame, degrees=True
    ):  # , active=_ACTIVE_DEFAULT):
        """
        getPassiveTranslationRotationVectorsTo(self,ref,degrees=True)

        extracts rotation vector from the result of getPassiveTransformationTo(target)
        """
        transformation = self.getPassiveTransformationTo(targetFrame)
        rotation = t3.euler.mat2euler(transformation, axes=self.rot_config)
        if degrees:
            rotation = np.array([np.rad2deg(item) for item in rotation])
        translation = transformation[:3, 3]
        return translation, rotation

    def getPassiveTranslationVectorTo(self, targetFrame, degrees=True):
        """
        getPassiveTranslationRotationVectorsTo(self,ref,degrees=True)

        extracts translation vector from the result of getPassiveTransformationTo(target)
        """
        return self.getPassiveTranslationRotationVectorsTo(targetFrame, degrees=degrees)[0]

    def getPassiveRotationVectorTo(self, targetFrame, degrees=True):
        """
        getPassiveTranslationRotationVectorsTo(self,targetFrame,degrees=True)

        extracts rotation vector from the result of getPassiveTransformationTo(targetFrame)
        """
        return self.getPassiveTranslationRotationVectorsTo(targetFrame, degrees=degrees)[1]

    def getPassiveTransformationFrom(self, source):
        """
        getPassiveTransformationFrom(self,source)
        ==  getPointTransformationTo(self,source)

        INPUT
        source : is a ReferenceFrame object

        OUTPUT
        returns the transformation matrix  that, applied to a point defined
        in source returns its coordinates in self
        """
        DEBUG and LOGGER.debug("PASSIVE FROM self {self.name} source {source.name}")
        return source.getPassiveTransformationTo(self)

    def getPassiveTranslationRotationVectorsFrom(
        self, source, degrees=True
    ):  # , active=_ACTIVE_DEFAULT):
        """
        getPassiveTranslationRotationVectorsFrom(self,source, degrees=True)

        extracts rotation vector from the result of getPassiveTransformationTo(target)
        """
        transformation = self.getPassiveTransformationFrom(source)
        rotation = t3.euler.mat2euler(transformation, axes=self.rot_config)
        if degrees:
            rotation = np.array([np.rad2deg(item) for item in rotation])
        translation = transformation[:3, 3]
        return translation, rotation

    def getPassiveTranslationVectorFrom(self, source, degrees=True):
        """
        getPassiveTranslationVectorFrom(self,source, degrees=True)

        extracts translation vector from the result of getPassiveTransformationFrom(source)
        """
        return self.getPassiveTranslationRotationVectorsFrom(source, degrees=degrees)[0]

    def getPassiveRotationVectorFrom(self, source, degrees=True):
        """
        getPassiveTranslationRotationVectorsFrom(self,source,degrees=True)

        extracts rotation vector from the result of getPassiveTransformationFrom(source)
        """
        return self.getPassiveTranslationRotationVectorsFrom(source, degrees=degrees)[1]

    def getActiveTransformationTo(self, target):
        """
        Return the transformation matrix from this reference frame (``self``) to
        the target reference frame.

        Applying this transformation to the ``self`` ReferenceFrame would render it
        identical to target.

        Applying this transformation to a point defined in ``self`` would move it to
        the same coordinates in target.

        :param target: is a ReferenceFrame object

        :return: the transformation matrix that defines the reference frame ``target``
            in ``self``, the current reference frame. The transformation from self
            to target

        """
        DEBUG and LOGGER.debug("ACTIVE TO self {self.name} target {target.name}")
        return target.getPassiveTransformationTo(self)

    def getActiveTranslationRotationVectorsTo(
        self, targetFrame, degrees=True
    ):  # ,active=_ACTIVE_DEFAULT):
        """
        getActiveTranslationRotationVectorsTo(self,ref,degrees=True)

        extracts rotation vector from the result of getActiveTransformationTo(target)
        """
        transformation = self.getActiveTransformationTo(targetFrame)
        rotation = t3.euler.mat2euler(transformation, axes=self.rot_config)
        if degrees:
            rotation = np.array([np.rad2deg(item) for item in rotation])
        translation = transformation[:3, 3]
        return translation, rotation

    def getActiveTranslationVectorTo(self, targetFrame, degrees=True):
        """
        getActiveTranslationRotationVectorsTo(self,ref,degrees=True)

        extracts translation vector from the result of getActiveTransformationTo(target)
        """
        return self.getActiveTranslationRotationVectorsTo(targetFrame, degrees=degrees)[0]

    def getActiveRotationVectorTo(self, targetFrame, degrees=True):
        """
        getActiveTranslationRotationVectorsTo(self,targetFrame,degrees=True)

        extracts rotation vector from the result of getActiveTransformationTo(targetFrame)
        """
        return self.getActiveTranslationRotationVectorsTo(targetFrame, degrees=degrees)[1]

    def getActiveTransformationFrom(self, source):
        """
        Applying this transformation to the ``source`` ReferenceFrame
        would render it identical to self.

        Applying this transformation to a point defined in source
        would move it to the same coordinates in self.

        :param source: is a ReferenceFrame object

        :return: the transformation matrix that defines this reference frame in ``source``,
            i.e. the transformation from ``source`` to ``self``

        """
        DEBUG and LOGGER.debug("ACTIVE FROM self {self.name} source {source.name}")
        return self.getPassiveTransformationTo(source)

    def getActiveTranslationRotationVectorsFrom(
        self, source, degrees=True
    ):  # ,active=_ACTIVE_DEFAULT):
        """
        getActiveTranslationRotationVectorsFrom(self,source, degrees=True)

        extracts rotation vector from the result of getActiveTransformationTo(target)
        """
        transformation = self.getActiveTransformationFrom(source)
        rotation = t3.euler.mat2euler(transformation, axes=self.rot_config)
        if degrees:
            rotation = np.array([np.rad2deg(item) for item in rotation])
        translation = transformation[:3, 3]
        return translation, rotation

    def getActiveTranslationVectorFrom(self, source, degrees=True):
        """
        getActiveTranslationVectorFrom(self,source, degrees=True)

        extracts translation vector from the result of getActiveTransformationFrom(source)
        """
        return self.getActiveTranslationRotationVectorsFrom(source, degrees=degrees)[0]

    def getActiveRotationVectorFrom(self, source, degrees=True):
        """
        getActiveTranslationRotationVectorsFrom(self,source,degrees=True)

        extracts rotation vector from the result of getActiveTransformationFrom(source)
        """
        return self.getActiveTranslationRotationVectorsFrom(source, degrees=degrees)[1]

    def _findEnds(self, frame, visited=[], ends=[], verbose=True, level=1):
        """
        PURPOSE
        Identify the 'linked_frames':
            frames that are linked, either directly or indirectly (via mult. links) to "frame"
            --> returned as visited
        Identify subset of 'linked_frames of which the reference does not belong to linked_frames
            --> returned as 'finalEnds'
        """
        DEBUG and LOGGER.debug(
            f"{level:-2d}{2*level*' '} Current: {frame.name} --  ends: {[f.name for f in ends]} -- visited {[f.name for f in visited]}"
        )
        # if verbose: print (f"{level:-2d}{2*level*' '} Current: {frame.name} --  ends: {[f.name for f in ends]} -- visited {[f.name for f in visited]}")

        # Establish the set of 'linked_frames' (variable 'visited')
        # The recursive process below keeps unwanted (non-endFrames), namely the
        # frames that are not directly, but well indirectly linked to their reference
        # This case is solved further down
        if frame not in visited:
            visited.append(frame)
            if verbose and level:
                level += 1
            if frame.ref not in frame.linkedTo:
                ends.append(frame)
                DEBUG and LOGGER.debug(f"{(10+2*level)*' '}{frame.name}: new end")
                # if verbose: LOGGER.info(f"{(10+2*level)*' '}{frame.name}: new end")
            for linkedFrame in frame.linkedTo:
                ends, visited = self._findEnds(
                    linkedFrame, visited=visited, ends=ends, verbose=verbose, level=level
                )

        # If frame.ref was linked to frame via an indirect route, reject it
        finalEnds = []
        for aframe in ends:
            if aframe.ref not in ends:
                finalEnds.append(aframe)
        return finalEnds, visited

    def setTransformation(
        self, transformation, updated=None, preserveLinks=True, _relative=False, verbose=True
    ):
        """
        setTransformation(self,transformation, updated=None, preserveLinks=True,_relative=False, verbose=True)

        Alter the definition of this coordinate system

        If other systems are linked to this one, their definition must be updated accordingly
          The link set between two ref. Frames A & B is the active transformation matrix from A to B
          A.addLink(B, matrix)
          A.getActiveTransformationTo(B) --> matrix

        The way to update the definition of the present system, and of those linked to it
        depends on the structure of those links.

        We define
        - the target frame as the one we want to move / redefine
        - 'linkedFrames' as those directly, or indirectly (i.e. via multiple links)
           linked to the target frame
        - endFrames as the subset of linkedFrames which are not linked to their reference (directly or indirectly)
        - sideFrames as the set of frames whose reference is a linkedFrame, but not themselves belonging to the linkedFrames

        We can demonstrate that updating the endFrames (Block A below) is sufficient to represent
        the movement of the target frame and all frames directly or indirectly linked to it.

        This may nevertheless have perverse effects for sideFrames. Indeed,
        their reference will (directly or implicitely) be redefined, but they shouldn't:
        they are not linked to their reference --> their location in space (e.g. wrt the master frame)
        should not be affected by the movement of the target frame. This is the aim of block B.

        For a completely robust solution, 2 steps must be taken
        BLOCK A. apply the rigt transformation to all "endFrames"
        BLOCK B. Check for frames
                       using any of the "visited" frames as a reference
                       not linked to its reference
            Correct its so that it doesn't move (it shouldn't be affected by the requested movement)
            This demands a "referenceFor" array property

        """
        # Ruthless, enforced redefinition of one system. Know what you do, or stay away.
        # Semi-unpredictible side effets if the impacted frame has links!
        if preserveLinks == False:
            self.transformation = transformation
            return

        if updated is None:
            updated = []

        # visitedFrames = all frames which can be reached from self via invariant links
        # endFrames = subset of visitedFrames that are at the end of a chain, and must be updated
        #             in order to properly represent the requested movement
        endFrames, visitedFrames = self._findEnds(frame=self, visited=[], ends=[], verbose=verbose)
        if verbose:
            LOGGER.info(
                f"Visited sub-system                      {[f.name for f in visitedFrames]}"
            )
            LOGGER.info(f"End-frames (movement necessary)         {[f.name for f in endFrames]}")

        # All updates are done by relative movements
        # so we must first compute the relative movement corresponding to the requested absolute movement
        if _relative == False:
            ## virtual = what self should become after the (absolute) movement
            ## it allows to compute the relative transformation to be applied and work in relative further down
            virtual = ReferenceFrame(
                transformation, ref=self.ref, name="virtual", rot_config=self.rot_config
            )
            request = self.getActiveTransformationTo(virtual)
            del virtual
        else:
            # If this method is called by applyTransformation,
            # we are facing a request for a relative movement
            # In that case the input is directly what we want
            request = transformation

        # BLOCK B. Check for frames that were impacted but shouldn't have been and correct them
        # B1. List of frames demanding a correction
        #     'impacted' are frames having their reference inside the rigid structure moving, but not linked to it
        #     If nothing is done, the movement will implicitely displace them, which is not intended

        ### Impacted shall not contain frames that are linked to self (== to any frame in visitedFrames) via any route...
        ### We check if the impacted frames are in visitedFrames:
        ### it is enough to know it's connected to the entire 'solid body' in which self belongs
        impacted = []
        for frame in visitedFrames:
            for child in frame.referenceFor:
                # Version 1 : too simple (restores too many frames)
                #if child not in frame.linkedTo:

                # Version 2 : overkill
                #child_ends, child_visited = child._findEnds(frame=child,visited=[],ends=[],verbose=verbose)
                #if frame not in child_visited:

                # Version 3 : just check if the child belongs to the rigid structure...
                if child not in visitedFrames:
                    impacted.append(child)

        DEBUG and LOGGER.debug(f"Impacted (not moving, defined in moving) {[f.name for f in impacted]}")

        # B2. save the location of all impacted frames
        # tempReference has the only purpose of avoiding that every frame must know the master
        # It could be any frame without links and defined wrt the master, but the master isn't known here...
        # TODO : confirm that the master isn't known (e.g. via cls._MASTER)
        tempMaster = self.findMaster()
        toRestore = {}
        for frame in impacted:
            toRestore[frame] = ReferenceFrame(
                frame.getActiveTransformationFrom(tempMaster),
                ref=tempMaster,
                name=frame.name + "toRestore",
                rot_config=frame.rot_config,
            )

        # BLOCK A. apply the rigt transformation to all "endFrames"
        """
        ### Ensure that "Untouched" remains unaffected regardless of the update order of the endFrames
        selfUntouched = ReferenceFrame(
            transformation = self.getActiveTransformationFrom(tempMaster),
            ref=tempMaster,
            name=self.name + "_fixed",
            rot_config=self.rot_config,
        )
        """

        selfUntouched = ReferenceFrame(
            transformation=self.transformation,
            ref=self.ref,
            name=self.name + "_fixed",
            rot_config=self.rot_config,
        )

        for bottom in endFrames:

            up = bottom.getActiveTransformationTo(selfUntouched)
            down = selfUntouched.getActiveTransformationTo(bottom)

            relativeTransformation = up @ request @ down

            if DEBUG:
                LOGGER.debug(
                    f"\nAdjusting {bottom.name} to {self.name}\nUpdated {[i.name for i in updated]}"
                )
                LOGGER.debug(f"\ninput transformation \n{np.round(transformation,3)}")
                LOGGER.debug(
                    f"\nup \n{np.round(up,3)}\ntransformation\n{np.round(request,3)}\ndown\n{np.round(down,3)}"
                )
                LOGGER.debug(f"\nrelativeTransformation \n{np.round(relativeTransformation,3)}")

            bottom.transformation = bottom.transformation @ relativeTransformation

            updated.append(bottom)

        for frame in visitedFrames:
            if frame not in updated:
                updated.append(frame)

        # Block B
        # B3. Correction
        # we must set preserveLinks to False to prevent cascading impact from this update
        # if X1 is impacted with
        #    X1.ref = E1     X1 --> X2 (simple link)    E2.ref = X2
        # where X1 and X2 are "external frames" and E1 and E2 are "endFrames" that will hence move
        # X1 was impacted by the move of E1, but X2 wasn't
        # ==> wrt master, neither X1 nor X2 should have moved, but X1 did (via its ref)
        # and hence its link with X2 is now corrupt
        # We need to move X1 back to its original location wrt master
        # if we preserved the links while doing that,
        # we will move X2, which shouldn't move
        # (it didn't have to, it didn't and the goal is to restore the validity of the links)
        #
        # Direct restoration or the impacted frames at their original location

        for frame in toRestore:
            frame.transformation = frame.ref.getActiveTransformationTo(toRestore[frame])

        del toRestore

        return

    def setTranslationRotation(
        self,
        translation,
        rotation,
        rot_config=_ROT_CONFIG_DEFAULT,
        active=_ACTIVE_DEFAULT,
        degrees=True,
        preserveLinks=True,
    ):
        """
        setTranslationRotation(self,translation,rotation,rot_config=_ROT_CONFIG_DEFAULT, active=_ACTIVE_DEFAULT, degrees=True, preserveLinks=True)

        Same as setTransformation, but input = translation and rotation vectors rather than affine transformation matrix
        """
        # Translation
        translation = np.array(translation)
        # Zoom - unit
        zdef = np.array([1, 1, 1])
        # Shear
        sdef = np.array([0, 0, 0])
        # Rotation
        if degrees:
            rotation = np.array([np.deg2rad(item) for item in rotation])

        rotx, roty, rotz = rotation

        rmat = RotationMatrix(rotx, roty, rotz, rot_config=rot_config, active=active)

        DEBUG and LOGGER.debug(t3.affines.compose(translation, rmat.R, Z=zdef, S=sdef))

        transformation = t3.affines.compose(translation, rmat.R, Z=zdef, S=sdef)

        self.setTransformation(transformation, preserveLinks=preserveLinks, _relative=False)
        return

    def applyTransformation(self, transformation, updated=None, preserveLinks=True):
        """
        applyTransformation(self,transformation)

        Applies the transformation to the current reference frame's definition

        self.transformation := transformation @ self.transformation
        """
        if updated is None:
            updated = []

        self.setTransformation(
            transformation=transformation,
            updated=updated,
            preserveLinks=preserveLinks,
            _relative=True,
        )

    def applyTranslationRotation(
        self,
        translation,
        rotation,
        rot_config=None,
        active=_ACTIVE_DEFAULT,
        degrees=True,
        preserveLinks=True,
    ):
        """
        applyTranslationRotation(self,translation,rotation,rot_config=None,active=_ACTIVE_DEFAULT,degrees=True, preserveLinks=True)

        Builds transformation from input translation and rotation, then applies this transformation
        """
        if rot_config is None:
            rot_config = self.rot_config

        # Translation
        translation = np.array(translation)
        # Zoom - unit
        zdef = np.array([1, 1, 1])
        # Shear
        sdef = np.array([0, 0, 0])
        # Rotation
        if degrees:
            rotation = np.array([np.deg2rad(item) for item in rotation])
        rotx, roty, rotz = rotation
        #
        rmat = RotationMatrix(rotx, roty, rotz, rot_config=rot_config, active=active)
        #
        transformation = t3.affines.compose(translation, rmat.R, Z=zdef, S=sdef)
        #
        self.applyTransformation(transformation, preserveLinks=preserveLinks)
        return

    def getAxis(self, axis, name=None):
        """
        INPUT
        axis : in ['x','y','z']

        OUTPUT
        Returns an object of class Point, corresponding to the vector defining
        the axis of choice in self.
        The output can be used with the Point methods to express that axis in any reference frame
        """
        haxis = {}
        haxis["x"] = [1, 0, 0]
        haxis["y"] = [0, 1, 0]
        haxis["z"] = [0, 0, 1]
        if name is None:
            name = self.name + axis
        return Point(haxis[axis], ref=self, name=name)

    def getNormal(self, name=None):
        """
        getNormal(self,name=None)

        Returns a unit vector normal to the X-Y plane (= [0,0,1] = getAxis('z'))
        """
        return Point([0, 0, 1], ref=self, name=name)

    def getOrigin(self, name=None):
        """
        OUTPUT
        Returns an object of class Point, corresponding to the vector defining the origin in 'self', i.e. [0,0,0]
        The output can be used with the Point methods to express that axis in any reference frame
        """
        return Point([0, 0, 0], ref=self, name=name)

    def is_master(self):
        return self.isMaster()

    def isMaster(self):
        """
        A Master reference frame is a reference frame that refers to itself, and the
        transformation is the Identity matrix.
        """
        tr = self.transformation

        if (
            (self.name == self.ref.name)
            and (tr.shape[0] == tr.shape[1])
            and np.allclose(tr, np.eye(tr.shape[0]))
        ):
            return True
        return False

    def isSame(self, other):
        """
        Returns True if the two reference frames are the same except for their name.

        Two Reference Frames are considered the same when:

        * their transformation matrices are equal
        * the reference frame is equal
        * the rot_config is equal
        * the name may be different

        .. todo:: This needs further work and testing!
        """

        if other is self:
            DEBUG and LOGGER.debug(
                "self and other are the same object (beware: this message might occur with recursion from self.ref != self.other)"
            )
            return True

        if isinstance(other, ReferenceFrame):
            DEBUG and LOGGER.debug(f"comparing {self.name} and {other.name}")
            if not np.array_equal(self.transformation, other.transformation):
                DEBUG and LOGGER.debug("self.transformation not equals other.transformation")
                return False
            if self.rot_config != other.rot_config:
                DEBUG and LOGGER.debug("self.rot_config not equals other.rot_config")
                return False
            # The following tests are here to prevent recursion to go infinite when self and other
            # point to itself
            if self.ref is self and other.ref is other:
                DEBUG and LOGGER.debug("both self.ref and other.ref point to themselves")
                pass
            else:
                DEBUG and LOGGER.debug("one of self.ref or other.ref doesn't points to itself")
                if self.ref != other.ref:
                    DEBUG and LOGGER.debug("self.ref not equals other.ref")
                    return False
            if self.name is not other.name:
                DEBUG and LOGGER.debug(
                    f"When checking two reference frames for equality, only their names differ: '{self.name}' not equals '{other.name}'"
                )
                pass
            return True
        return NotImplemented

    def __eq__(self, other):
        """
        Overrides the default implementation, which basically checks for id(self) == id(other).

        Two Reference Frames are considered equal when:

        * their transformation matrices are equal
        * the reference frame is equal
        * the rot_config is equal
        * do we want to insist on the name being equal?
          YES - for strict testing
          NO  - this might need a new method like isSame(self, other) where the criteria are relaxed


        .. todo:: This needs further work and testing!
        """

        if other is self:
            DEBUG and LOGGER.debug(
                "self and other are the same object (beware: this message might occur with recursion from self.ref != self.other)"
            )
            return True

        if isinstance(other, ReferenceFrame):
            DEBUG and LOGGER.debug(f"comparing {self.name} and {other.name}")
            if not np.array_equal(self.transformation, other.transformation):
                DEBUG and LOGGER.debug("self.transformation not equals other.transformation")
                return False
            if self.rot_config != other.rot_config:
                DEBUG and LOGGER.debug("self.rot_config not equals other.rot_config")
                return False
            # The following tests are here to prevent recursion to go infinite when self and other
            # point to itself
            if self.ref is self and other.ref is other:
                DEBUG and LOGGER.debug("both self.ref and other.ref point to themselves")
                pass
            else:
                DEBUG and LOGGER.debug("one of self.ref or other.ref doesn't points to itself")
                if self.ref != other.ref:
                    DEBUG and LOGGER.debug("self.ref not equals other.ref")
                    return False
            if self.name is not other.name:
                DEBUG and LOGGER.debug(
                    f"When checking two reference frames for equality, only their names differ: '{self.name}' not equals '{other.name}'"
                )
                return False
            return True
        return NotImplemented

    def __hash__(self):
        """Overrides the default implementation"""
        hash_number = (id(self.rot_config) + id(self.ref) + id(self.name)) // 16
        return hash_number

    def __copy__(self):

        DEBUG and LOGGER.debug(
            f'Copying {self!r} unless {self.name} is "Master" in which case the Master itself is returned.'
        )

        if self.isMaster():
            DEBUG and LOGGER.debug(f"Returning Master itself instead of a copy.")
            return self

        return ReferenceFrame(self.transformation, self.ref, self.name, self.rot_config)



import logging
from pathlib import Path

from egse.command import ClientServerCommand
from egse.control import ControlServer
from egse.device import DeviceConnectionState
from egse.hexapod.symetrie import ControllerFactory
from egse.hexapod.symetrie.puna import PunaInterface
from egse.hexapod.symetrie.puna import PunaSimulator
from egse.hexapod.symetrie import get_hexapod_controller_pars
# from egse.hk import read_conversion_dict, convert_hk_names
from egse.protocol import CommandProtocol
from egse.settings import Settings
from egse.system import format_datetime
from egse.zmq_ser import bind_address

_LOGGER = logging.getLogger(__name__)

_HERE = Path(__file__).parent

ctrl_settings = Settings.load("Hexapod PUNA Control Server")
PUNA_SETTINGS = Settings.load("PMAC Controller")
DEVICE_SETTINGS = Settings.load(filename="puna.yaml", location=_HERE)


class PunaCommand(ClientServerCommand):
    pass


class PunaProtocol(CommandProtocol):
    def __init__(self, control_server: ControlServer, simulator: bool = False):
        super().__init__()
        self.control_server = control_server
        self.simulator = simulator

        # self.hk_conversion_table = read_conversion_dict(self.control_server.get_storage_mnemonic(), use_site=True)

        if self.simulator:
            self.hexapod = PunaSimulator()
        else:
            hostname, port, device_id, device_name, controller_type = get_hexapod_controller_pars()

            factory = ControllerFactory()
            self.hexapod = factory.create(device_name, device_id=device_id)
            # self.hexapod = PunaController(hostname=hostname, port=port)
            self.hexapod.add_observer(self)

        try:
            self.hexapod.connect()
        except ConnectionError:
            _LOGGER.warning("Couldn't establish a connection to the PUNA Hexapod, check the log messages.")

        self.load_commands(DEVICE_SETTINGS.Commands, PunaCommand, PunaInterface)
        self.build_device_method_lookup_table(self.hexapod)

    def get_bind_address(self):
        return bind_address(
            self.control_server.get_communication_protocol(),
            self.control_server.get_commanding_port(),
        )

    def get_device(self):
        return self.hexapod

    def get_status(self):

        status = super().get_status()

        if self.state == DeviceConnectionState.DEVICE_NOT_CONNECTED and not self.simulator:
            return status

        mach_positions = self.hexapod.get_machine_positions()
        user_positions = self.hexapod.get_user_positions()
        actuator_length = self.hexapod.get_actuator_length()

        status.update({"mach": mach_positions, "user": user_positions, "alength": actuator_length})

        return status

    def get_housekeeping(self) -> dict:

        result = dict()
        result["timestamp"] = format_datetime()

        if self.state == DeviceConnectionState.DEVICE_NOT_CONNECTED and not self.simulator:
            return result

        mach_positions = self.hexapod.get_machine_positions()
        user_positions = self.hexapod.get_user_positions()
        actuator_length = self.hexapod.get_actuator_length()

        # The result of the previous calls might be None when e.g. the connection
        # to the device gets lost.

        if mach_positions is None or user_positions is None or actuator_length is None:
            if not self.hexapod.is_connected():
                _LOGGER.warning("Hexapod PUNA disconnected.")
                self.update_connection_state(DeviceConnectionState.DEVICE_NOT_CONNECTED)
            return result

        for idx, key in enumerate(
            ["user_t_x", "user_t_y", "user_t_z", "user_r_x", "user_r_y", "user_r_z"]
        ):
            result[key] = user_positions[idx]

        for idx, key in enumerate(
            ["mach_t_x", "mach_t_y", "mach_t_z", "mach_r_x", "mach_r_y", "mach_r_z"]
        ):
            result[key] = mach_positions[idx]

        for idx, key in enumerate(
            ["alen_t_x", "alen_t_y", "alen_t_z", "alen_r_x", "alen_r_y", "alen_r_z"]
        ):
            result[key] = actuator_length[idx]

        # TODO:
        #   the get_general_state() method should be refactored as to return a dict instead of a
        #   list. Also, we might want to rethink the usefulness of returning the tuple,
        #   it the first return value ever used?

        _, _ = self.hexapod.get_general_state()

        result["Homing done"] = self.hexapod.is_homing_done()
        result["In position"] = self.hexapod.is_in_position()

        return result  # convert_hk_names(result, self.hk_conversion_table)

    def is_device_connected(self):
        # FIXME(rik): There must be another way to check if the socket is still alive...
        #             This will send way too many VERSION requests to the controllers.
        #             According to SO [https://stackoverflow.com/a/15175067] the best way
        #             to check for a connection drop / close is to handle the exceptions
        #             properly.... so, no polling for connections by sending it a simple
        #             command.
        return self.hexapod.is_connected()
